import typing

import cf_xarray
import metpy.interpolate
import numpy as np
import xarray as xr

supported_types = typing.Union[xr.DataArray, np.ndarray]

__pres_lev_mandatory__ = np.array([
    1000, 925, 850, 700, 500, 400, 300, 250, 200, 150, 100, 70, 50, 30, 20, 10,
    7, 5, 3, 2, 1
]).astype(np.float32)  # Mandatory pressure levels (mb)
__pres_lev_mandatory__ = __pres_lev_mandatory__ * 100.0  # Convert mb to Pa


def _func_interpolate(method='linear'):
    """Define interpolation function."""

    if method == 'linear':
        func_interpolate = metpy.interpolate.interpolate_1d
    elif method == 'log':
        func_interpolate = metpy.interpolate.log_interpolate_1d
    else:
        raise ValueError(f'Unknown interpolation method: {method}. '
                         f'Supported methods are: "log" and "linear".')

    return func_interpolate


def _pressure_from_hybrid(psfc, hya, hyb, p0=100000.):
    """Calculate pressure at the hybrid levels."""

    # p(k) = hya(k) * p0 + hyb(k) * psfc

    # This will be in Pa
    return hya * p0 + hyb * psfc


def _pre_interp_multidim(
    data_in: xr.DataArray,
    cyclic: bool,
    missing_val,
):
    """Helper Function: Handling missing data functionality and adding cyclic
    point if required.

    Parameters
    ----------
    data_in : :class:`xarray.DataArray`
        The data on which to operate

    cyclic : :class:`bool`
        Determines if cyclic point should be added or not.
        If true then add point, else do nothing.

    missing_val : int, float, optional
        Provides an alternative to NaN

    Returns
    -------
    data_in : :class:`xarray.DataArray`
       The data input with cyclic points added (if cyclic is true)
       and missing_val values replaced with np.nan

    Notes
    -------
    """
    # replace missing_val with np.nan
    if missing_val is not None:
        data_in = xr.DataArray(np.where(data_in.values == missing_val, np.nan,
                                        data_in.values),
                               dims=data_in.dims,
                               coords=data_in.coords)

    # add cyclic points and create new data array
    if cyclic:
        padded_data = np.pad(data_in.values, ((0, 0), (1, 1)), mode='wrap')
        padded_longitudes = np.pad(data_in.coords[data_in.dims[-1]], (1, 1),
                                   mode='wrap')
        padded_longitudes[0] -= 360
        padded_longitudes[-1] += 360

        data_in = xr.DataArray(
            padded_data,
            coords={
                data_in.dims[-2]: data_in.coords[data_in.dims[-2]].values,
                data_in.dims[-1]: padded_longitudes,
            },
            dims=data_in.dims,
        )

    return data_in


def _post_interp_multidim(data_in, missing_val):
    """Helper Function: Handling missing data functionality.

    Parameters
    ----------
    data_in : :class:`xarray.DataArray`
        The data on which to operate

    missing_val : int, float, optional
         Provides an alternative to NaN

    Returns
    -------
    data_in : :class:`xarray.DataArray`
       The data input with np.nan values replaced with missing_val
    """
    if missing_val is not None:
        data_in = xr.DataArray(np.where(np.isnan(data_in.values), missing_val,
                                        data_in.values),
                               dims=data_in.dims,
                               coords=data_in.coords)

    return data_in


def _sigma_from_hybrid(psfc, hya, hyb, p0=100000.):
    """Calculate sigma at the hybrid levels."""

    # sig(k) = hya(k) * p0 / psfc + hyb(k)

    # This will be in Pa
    return hya * p0 / psfc + hyb


def _vertical_remap(func_interpolate, new_levels, xcoords, data, interp_axis=0):
    """Execute the defined interpolation function on data."""

    return func_interpolate(new_levels, xcoords, data, axis=interp_axis)


def _temp_extrapolate(data, lev_dim, lev, p_sfc, ps, phi_sfc):
    r"""This helper function extrapolates temperature below ground using the
    ECMWF formulation described in `Vertical Interpolation and Truncation of
    Model-Coordinate Data <http://dx.doi.org/10.5065/D6HX19NH>`__ by Trenberth,
    Berry, & Buja [NCAR/TN-396, 1993]. Specifically equation 16 is used:

    .. math::
        T = T_* \left( 1 + \alpha ln \frac{p}{p_s} + \frac{1}{2}\left( \alpha ln \frac{p}{p_s} \right)^2 + \frac{1}{6} \left( \alpha ln \frac{p}{p_s} \right)^3 \right)

    Parameters
    ----------
    data: :class:`xarray.DataArray`
        The temperature at the lowest level of the model.

    lev_dim: str
        The name of the vertical dimension.

    lev: int
        The pressure levels of interest. Must be in the same units as ``ps`` and ``p_sfc``

    p_sfc: :class:`xarray.DataArray`
        The pressure at the lowest level of the model. Must be in the same units as ``lev`` and ``ps``

    ps: :class:`xarray.DataArray`
        An array of surface pressures. Must be in the same units as ``lev`` and ``p_sfc``

    phi_sfc: :class:`xarray.DataArray`
        The geopotential at the lowest level of the model.

    Returns
    -------
    result: :class:`xarray.DataArray`
        The extrapolated temepratures at the provided pressure levels.
    """
    R_d = 287.04  # dry air gas constant
    g_inv = 1 / 9.80616  # inverse of gravity
    alpha = 0.0065 * R_d * g_inv

    tstar = data.isel(**{lev_dim: -1}) * (1 + alpha * (ps / p_sfc - 1))
    hgt = phi_sfc * g_inv
    t0 = tstar + 0.0065 * hgt
    tplat = xr.apply_ufunc(np.minimum, 298, t0, dask='parallelized')

    tprime0 = xr.where((2000 <= hgt) & (hgt <= 2500),
                       0.002 * ((2500 - hgt) * t0 + ((hgt - 2000) * tplat)),
                       np.nan)
    tprime0 = xr.where(2500 < hgt, tplat, np.nan)

    alnp = xr.where(hgt < 2000, alpha * np.log(lev / ps),
                    R_d * (tprime0 - tstar) / phi_sfc * np.log(lev / ps))
    alnp = xr.where(tprime0 < tstar, 0, alnp)

    return tstar * (1 + alnp + (0.5 * (alnp**2)) + (1 / 6 * (alnp**3)))


def _geo_height_extrapolate(t_bot, lev, p_sfc, ps, phi_sfc):
    r"""This helper function extrapolates geopotential height below ground using
    the ECMWF formulation described in `Vertical Interpolation and Truncation
    of Model-Coordinate Data <http://dx.doi.org/10.5065/D6HX19NH>`__ by
    Trenberth, Berry, & Buja [NCAR/TN-396, 1993]. Specifically equation 15 is
    used:

    .. math::
        \Phi = \Phi_s - R_d T_* ln \frac{p}{p_s} \left[ 1 + \frac{1}{2}\alpha ln\frac{p}{p_s} + \frac{1}{6} \left( \alpha ln \frac{p}{p_s} \right)^2 \right]

    Parameters
    ----------
    t_bot: :class:`xarray.DataArray`
        Temperature at the lowest (bottom) level of the model.

    lev: int
        The pressure level of interest. Must be in the same units as ``ps`` and ``p_sfc``

    p_sfc: :class:`xarray.DataArray`
        The pressure at the lowest level of the model. Must be in the same units as ``lev`` and ``ps``

    ps : :class:`xarray.DataArray`
        An array of surface pressures. Must be in the same units as ``lev`` and ``p_sfc``

    phi_sfc:
        The geopotential at the lowest level of the model.

    Returns
    -------
    result: :class:`xarray.DataArray`
        The extrapolated geopotential height in geopotential meters at the provided pressure levels.
    """
    R_d = 287.04  # dry air gas constant
    g_inv = 1 / 9.80616  # inverse of gravity
    alpha = 0.0065 * R_d * g_inv

    tstar = t_bot * (1 + alpha * (ps / p_sfc - 1))
    hgt = phi_sfc * g_inv
    t0 = tstar + 0.0065 * hgt

    alph = xr.where((tstar <= 290.5) & (t0 > 290.5),
                    R_d / phi_sfc * (290.5 - tstar), alpha)

    alph = xr.where((tstar > 290.5) & (t0 > 290.5), 0, alph)
    tstar = xr.where((tstar > 290.5) & (t0 > 290.5), 0.5 * (290.5 + tstar),
                     tstar)

    tstar = xr.where((tstar < 255), 0.5 * (tstar + 255), tstar)

    alnp = alph * np.log(lev / ps)
    return hgt - R_d * tstar * g_inv * np.log(
        lev / ps) * (1 + 0.5 * alnp + 1 / 6 * alnp**2)


def _vertical_remap_extrap(new_levels, lev_dim, data, output, pressure, ps,
                           variable, t_bot, phi_sfc):
    """A helper function to call the appropriate extrapolation function based
    on the user's inputs.

    Parameters
    ----------
    new_levels: array-like
        The desired pressure levels for extrapolation in Pascals.

    lev_dim: str
        The name of the vertical dimension.

    data: :class:`xarray.DataArray`
        The data to extrapolate

    output: :class:`xarray.DataArray`
        An array to hold the output data

    pressure: :class:`xarray.DataArray`
        The pressure at the lowest level of the model. Must be in the same units as ``lev`` and ``ps``

    ps : :class:`xarray.DataArray`
        An array of surface pressures. Must be in the same units as ``lev`` and ``p_sfc``

    variable : str, optional
        String representing what variable is extrapolated below surface level.
        Temperature extrapolation = "temperature". Geopotential height
        extrapolation = "geopotential". All other variables = "other". If
        "other", the value of ``data`` at the lowest model level will be used
        as the below ground fill value. Required if extrapolate is True.

    t_bot: :class:`xarray.DataArray`
        Temperature at the lowest (bottom) level of the model.

    phi_sfc:
        The geopotential at the lowest level of the model.

    Returns
    -------
    output: :class:`xarray.DataArray`
        A DataArray containing the data after extrapolation.
    """
    sfc_index = pressure[lev_dim].argmax().data  # index of the model surface
    p_sfc = pressure.isel(**dict({lev_dim: sfc_index
                                 }))  # extract pressure at lowest level

    if variable == 'temperature':
        for lev in new_levels:
            output.loc[dict(plev=lev)] = xr.where(
                lev <= p_sfc, output.sel(plev=lev),
                _temp_extrapolate(data, lev_dim, lev, p_sfc, ps, phi_sfc))

    elif variable == 'geopotential':
        for lev in new_levels:
            output.loc[dict(plev=lev)] = xr.where(
                lev <= p_sfc, output.sel(plev=lev),
                _geo_height_extrapolate(t_bot, lev, p_sfc, ps, phi_sfc))
    else:
        for lev in new_levels:
            output.loc[dict(plev=lev)] = xr.where(
                lev <= p_sfc, output.sel(plev=lev),
                data.isel(**dict({lev_dim: sfc_index})))
    return output


def interp_hybrid_to_pressure(data: xr.DataArray,
                              ps: xr.DataArray,
                              hyam: xr.DataArray,
                              hybm: xr.DataArray,
                              p0: float = 100000.,
                              new_levels: np.ndarray = __pres_lev_mandatory__,
                              lev_dim: str = None,
                              method: str = 'linear',
                              extrapolate: bool = False,
                              variable: str = None,
                              t_bot: xr.DataArray = None,
                              phi_sfc: xr.DataArray = None) -> xr.DataArray:
    """Interpolate and extrapolate data from hybrid-sigma levels to isobaric
    levels. Keeps attributes (i.e. metadata) of the input data in the output as
    default.

    Notes
    -----
    ACKNOWLEDGEMENT: We'd like to thank to `Brian Medeiros <https://github.com/brianpm>`__,
    `Matthew Long <https://github.com/matt-long>`__, and `Deepak Cherian <https://github.com/dcherian>`__
    at NCAR for their great contributions since the code implemented here is mostly
    based on their work.

    Parameters
    ----------
    data : :class:`xarray.DataArray`
        Multidimensional data array of hybrid-sigma levels and has a ``lev_dim`` coordinate.

    ps : :class:`xarray.DataArray`
        A multi-dimensional array of surface pressures (Pa), same time/space shape as data.

    hyam, hybm : :class:`xarray.DataArray`
        One-dimensional arrays containing the hybrid A and B coefficients. Must have the same
        dimension size as the ``lev_dim`` dimension of data.

    p0 : float, optional
        Scalar numeric value equal to surface reference pressure (Pa). Defaults to 100000 Pa.

    new_levels : ndarray, optional
        A one-dimensional array of output pressure levels (Pa). If not given, the mandatory
        list of 21 pressure levels is used.

    lev_dim : str, optional
        String that is the name of level dimension in data. Defaults to "lev".

    method : str, optional
        String that is the interpolation method; can be either "linear" or "log". Defaults to "linear".

    extrapolate : bool, optional
        If True, below ground extrapolation for ``variable`` will be done using
        an `ECMWF formulation <http://dx.doi.org/10.5065/D6HX19NH>`__. Defaults
        to False.

    variable : str, optional
        String representing what variable is extrapolated below surface level.
        Temperature extrapolation = "temperature". Geopotential height
        extrapolation = "geopotential". All other variables = "other". If
        "other", the value of ``data`` at the lowest model level will be used
        as the below ground fill value. Required if extrapolate is True.

    t_bot : :class:`xarray.DataArray`, optional
        Temperature in Kelvin at the lowest layer of the model. Not necessarily
        the same as surface temperature. Required if ``extrapolate`` is True
        and ``variable`` is not ``'other'``

    phi_sfc: :class:`xarray.DataArray`, optional
        Geopotential in J/kg at the lowest layer of the model. Not necessarily
        the same as surface geopotential. Required if ``extrapolate`` is True
        and ``variable`` is not ``'other'``.

    Returns
    -------
    output : :class:`xarray.DataArray`
        Interpolated data with isobaric levels

    See Also
    --------
    Related NCL Functions:
    `vinth2p <https://www.ncl.ucar.edu/Document/Functions/Built-in/vinth2p.shtml>`__,
    `vinth2p_ecmwf <https://www.ncl.ucar.edu/Document/Functions/Built-in/vinth2p_ecmwf.shtml>`__
    """

    # Check inputs
    if (extrapolate and (variable is None)):
        raise ValueError(
            "If `extrapolate` is True, `variable` must be provided.")

    if variable in ['geopotential', 'temperature'] and (t_bot is None or
                                                        phi_sfc is None):
        raise ValueError(
            "If `variable` is 'geopotential' or 'temperature', both `t_bot` and `phi_sfc` must be provided"
        )

    if (variable not in ['geopotential', 'temperature', 'other', None]):
        raise ValueError(
            "The value of `variable` is " + variable +
            ", but the accepted values are 'temperature', 'geopotential', 'other', or None."
        )

    # Determine the level dimension and then the interpolation axis
    if lev_dim is None:
        try:
            lev_dim = data.cf["vertical"].name
        except Exception:
            raise ValueError(
                "Unable to determine vertical dimension name. Please specify the name via `lev_dim` argument."
            )

    try:
        func_interpolate = _func_interpolate(method)
    except ValueError as vexc:
        raise ValueError(vexc.args[0])

    interp_axis = data.dims.index(lev_dim)

    # Calculate pressure levels at the hybrid levels
    pressure = _pressure_from_hybrid(ps, hyam, hybm, p0)  # Pa

    # Make pressure shape same as data shape
    pressure = pressure.transpose(*data.dims)

    ###############################################################################
    # Workaround
    #
    # For the issue with metpy's xarray interface:
    #
    # `metpy.interpolate.interpolate_1d` had "no implementation found for
    # 'numpy.apply_along_axis'" issue for cases where the input is
    # xarray.Dataarray and has more than 3 dimensions (e.g. 4th dim of `time`).

    # Use dask.array.core.map_blocks instead of xarray.apply_ufunc and
    # auto-chunk input arrays to ensure using only Numpy interface of
    # `metpy.interpolate.interpolate_1d`.

    # # Apply vertical interpolation
    # # Apply Dask parallelization with xarray.apply_ufunc
    # output = xr.apply_ufunc(
    #     _vertical_remap,
    #     data,
    #     pressure,
    #     exclude_dims=set((lev_dim,)),  # Set dimensions allowed to change size
    #     input_core_dims=[[lev_dim], [lev_dim]],  # Set core dimensions
    #     output_core_dims=[["plev"]],  # Specify output dimensions
    #     vectorize=True,  # loop over non-core dims
    #     dask="parallelized",  # Dask parallelization
    #     output_dtypes=[data.dtype],
    #     dask_gufunc_kwargs={"output_sizes": {
    #         "plev": len(new_levels)
    #     }},
    # )

    # If an unchunked Xarray input is given, chunk it just with its dims
    if data.chunks is None:
        data_chunk = dict([
            (k, v) for (k, v) in zip(list(data.dims), list(data.shape))
        ])
        data = data.chunk(data_chunk)

    # Chunk pressure equal to data's chunks
    pressure = pressure.chunk(data.chunks)

    # Output data structure elements
    out_chunks = list(data.chunks)
    out_chunks[interp_axis] = (new_levels.size,)
    out_chunks = tuple(out_chunks)
    # ''' end of boilerplate

    from dask.array.core import map_blocks
    output = map_blocks(
        _vertical_remap,
        func_interpolate,
        new_levels,
        pressure.data,
        data.data,
        interp_axis,
        chunks=out_chunks,
        dtype=data.dtype,
        drop_axis=[interp_axis],
        new_axis=[interp_axis],
    )

    # End of Workaround
    ###############################################################################

    output = xr.DataArray(output, name=data.name, attrs=data.attrs)

    # Set output dims and coords
    dims = [
        data.dims[i] if i != interp_axis else "plev" for i in range(data.ndim)
    ]

    # Rename output dims. This is only needed with above workaround block
    dims_dict = {output.dims[i]: dims[i] for i in range(len(output.dims))}
    output = output.rename(dims_dict)

    coords = {}
    for (k, v) in data.coords.items():
        if k != lev_dim:
            coords.update({k: v})
        else:
            coords.update({"plev": new_levels})

    output = output.transpose(*dims).assign_coords(coords)

    if extrapolate:
        output = _vertical_remap_extrap(new_levels, lev_dim, data, output,
                                        pressure, ps, variable, t_bot, phi_sfc)

    return output


def interp_sigma_to_hybrid(data: xr.DataArray,
                           sig_coords: xr.DataArray,
                           ps: xr.DataArray,
                           hyam: xr.DataArray,
                           hybm: xr.DataArray,
                           p0: float = 100000.,
                           lev_dim: str = None,
                           method: str = 'linear') -> xr.DataArray:
    """Interpolate data from sigma to hybrid coordinates.  Keeps the attributes
    (i.e. meta information) of the input data in the output as default.

    Parameters
    ----------
    data : :class:`xarray.DataArray`
        Multidimensional data array, which holds sigma levels and has a ``lev_dim`` coordinate.

    sig_coords : :class:`xarray.DataArray`
        A one-dimensional array of sigma coordinates of ``lev_dim`` of ``data``.

    ps : :class:`xarray.DataArray`
        A multi-dimensional array of surface pressures (Pa), same time/space shape as data.

    hyam, hybm : :class:`xarray.DataArray`
        One-dimensional arrays containing the hybrid A and B coefficients. Must have the same
        dimension as the output hybrid levels.

    p0 : float, optional
        Scalar numeric value equal to surface reference pressure (Pa). Defaults to 100000 Pa.

    lev_dim : str, optional
        String that is the name of level dimension in data. Defaults to "lev".

    method : str, optional
        String that is the interpolation method; can be either "linear" or "log". Defaults to "linear".

    Returns
    -------
    output : :class:`xarray.DataArray`
        Interpolated data with hybrid levels

    See Also
    --------
    Related NCL Function:
    `sigma2hybrid <https://www.ncl.ucar.edu/Document/Functions/Built-in/sigma2hybrid.shtml>`__
    """

    # Determine the level dimension and then the interpolation axis
    if lev_dim is None:
        try:
            lev_dim = data.cf["vertical"].name
        except Exception:
            raise ValueError(
                "Unable to determine vertical dimension name. Please specify the name via `lev_dim` argument.'"
            )

    try:
        func_interpolate = _func_interpolate(method)
    except ValueError as vexc:
        raise ValueError(vexc.args[0])

    # Calculate sigma levels at the hybrid levels
    sigma = _sigma_from_hybrid(ps, hyam, hybm, p0)  # Pa

    non_lev_dims = list(data.dims)
    if (data.ndim > 1):
        non_lev_dims.remove(lev_dim)
        data_stacked = data.stack(combined=non_lev_dims).transpose()
        sigma_stacked = sigma.stack(combined=non_lev_dims).transpose()

        h_coords = sigma_stacked[0, :].copy()

        output = data_stacked[:, :len(hyam)].copy()

        for idx, (d, s) in enumerate(zip(data_stacked, sigma_stacked)):
            output[idx, :] = xr.DataArray(
                _vertical_remap(func_interpolate, s.data, sig_coords.data,
                                d.data))

        # Make output shape same as data shape
        output = output.unstack().transpose(*data.dims)
    else:
        h_coords = sigma

        output = data[:len(hyam)].copy()
        output[:len(hyam)] = xr.DataArray(
            _vertical_remap(func_interpolate, sigma.data, sig_coords.data,
                            data.data))

    # Set output dims and coords
    output = output.rename({lev_dim: 'hlev'})
    output = output.assign_coords({"hlev": h_coords.data})

    return output


def interp_multidim(
        data_in: supported_types,
        lat_out: np.ndarray,
        lon_out: np.ndarray,
        lat_in: np.ndarray = None,
        lon_in: np.ndarray = None,
        cyclic: bool = False,
        missing_val: np.number = None,
        method: str = "linear",
        fill_value: typing.Union[str, np.number] = np.nan) -> supported_types:
    """Multidimensional interpolation of variables. Uses ``xarray.interp`` to
    perform linear interpolation. Will not perform extrapolation, returns
    missing values if any surrounding points contain missing values.

    Parameters
    ----------
    data_in : :class:`xarray.DataArray`, ndarray
        Data array with data to be interpolated and associated coords. If
        it is a np array, then ``lat_in`` and ``lon_in`` must be provided. Length must
        be coordinated with given coordinates.

    lat_out: ndarray
        List of latitude coordinates to be interpolated to.

    lon_out: ndarray
        List of longitude coordinates to be interpolated to.

    lat_in: ndarray
        List of latitude coordinates corresponding to ``data_in``. Must be
        given if ``data_in`` is not an xarray.

    lon_in: ndarray
        List of longitude coordinates corresponding to ``data_in``. Must be
        given if ``data_in`` is not an xarray.

    cyclic: bool, optional
        Set as true if lon values are cyclical but do not fully wrap around
        the globe
        (0, 1.5, 3, ..., 354, 355.5) Default is false

    missing_val : :class:`np.number`, optional
        Provide a number to represent missing data. Alternative to using ``np.nan``

    method: str, optional
        Provide specific method of interpolation. Default is "linear"
        “linear” or “nearest” for multidimensional array

    fill_value: str, optional
        Set as 'extrapolate' to allow extrapoltion of data. Default is
        no extrapolation.

    Returns
    -------
    data_out : ndarray, :class:`xarray.DataArray`
       Returns same data type as input ``data_in``. Shape will be the same as
       input array except
       for last two dimensions which will be equal to thecoordinates given in
       ``data_out``

    Examples
    --------
    >>> import xarray as xr
    >>> import numpy as np
    >>> import geocat.comp
    >>> data = np.asarray([[1, 2, 3, 4, 5, 99],
    ...                   [2, 4, 6, 8, 10, 12]])
    >>> lat_in = [0, 1]
    >>> lon_in = [0, 50, 100, 250, 300, 350]
    >>> data_in = xr.DataArray(data,
    ...                        dims=['lat', 'lon'],
    ...                        coords={'lat':lat_in,
    ...                                'lon': lon_in})
    >>> data_out = xr.DataArray(dims=['lat', 'lon'],
    ...                         coords={'lat': [0, 1],
    ...                                 'lon': [0, 50, 360]})
    >>> do = interp_multidim(data_in,
    ...                      [0, 1],
    ...                      [0, 50, 360],
    ...                      cyclic=True,
    ...                      missing_val=99)
    >>> print(do)
    <xarray.DataArray (lat: 2, lon: 3)>
    array([[ 1.,  2., 99.],
       [ 2.,  4., 99.]])
    Coordinates:
      * lat      (lat) int64 0 1
      * lon      (lon) int64 0 50 360

    See Also
    --------
    Related External Functions:
    `xarray.DataArray.interp <https://docs.xarray.dev/en/stable/generated/xarray.DataArray.interp.html>`__,
    `cartopy.util.add_cyclic_point <https://scitools.org.uk/cartopy/docs/latest/reference/generated/cartopy.util.add_cyclic_point.html>`__

    Related NCL Function:
    `NCL linint2 <https://www.ncl.ucar.edu/Document/Functions/Built-in/linint2.shtml>`__
    """
    # check for xarray/numpy
    if not isinstance(data_in, xr.DataArray):
        if lat_in is None or lon_in is None:
            raise ValueError(
                "Argument lat_in and lon_in must be provided if data_in is not an xarray"
            )
        data_in = xr.DataArray(data_in,
                               dims=['lat', 'lon'],
                               coords={
                                   'lat': lat_in,
                                   'lon': lon_in
                               })

    output_coords = {
        data_in.dims[-1]: lon_out,
        data_in.dims[-2]: lat_out,
    }

    data_in_modified = _pre_interp_multidim(data_in, cyclic, missing_val)
    data_out = data_in_modified.interp(output_coords,
                                       method=method,
                                       kwargs={'fill_value': fill_value})
    data_out_modified = _post_interp_multidim(data_out, missing_val=missing_val)

    return data_out_modified



from typing import Union

import numpy as np
import xarray as xr

SupportedTypes = Union[np.ndarray, xr.DataArray]
XTypes = Union[xr.DataArray, xr.Dataset]

d2r = 1.74532925199432957692369e-02  # degrees to radians conversion factor


def _rad_lat_wgs84(lat: SupportedTypes,):
    r"""The radius calculation for the wgs84 ellipsoid at a latitude uses a
    taylor series from.

    .. math::
        radius = \sqrt{a^2 \cdot cos(lat)^2+b^2 \cdot sin(lat)^2}

    This returns the radius of the ellipsoid for a given latitude
    This is accurate to within floating point error.

    Note
    ----
    This doesn't need to be a taylor series, though the taylor series
    is faster and a needed step for the _arc_lat_wgs84 function to avoid the
    elliptic integral

    Parameters
    ----------
    lat : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        2-dimensional dataset, of orthographic latitude coordinates

    Returns
    -------
    gradients : list of :class:`numpy.ndarray`, list of :class:`xarray.DataArray`
        latitudinal radii calculated using th WGS84 geoid.

    See Also
    --------
    Related NCL Functions:
    `grad_latlon_cfd <https://www.ncl.ucar.edu/Document/Functions/Contributed/grad_latlon_cfd.shtml>`__,
    `gradsf <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsf.shtml>`__,
    `gradsg <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsg.shtml>`__
    """

    return \
        8.05993093251779959604912e-107 * lat ** 48 - \
        1.26581811418535723456176e-102 * lat ** 46 - \
        9.10951565776720242021392e-98 * lat ** 44 + \
        7.49821836126201522491765e-93 * lat ** 42 - \
        2.27271626922827622448519e-88 * lat ** 40 - \
        2.71379394439952534826763e-84 * lat ** 38 + \
        6.14832871773468827219624e-79 * lat ** 36 - \
        2.84993185787053811467259e-74 * lat ** 34 + \
        3.04677525449067422497153e-70 * lat ** 32 + \
        4.32285294849656618043972e-65 * lat ** 30 - \
        3.09781963775152747156702e-60 * lat ** 28 + \
        8.20657704323096444142572e-56 * lat ** 26 + \
        2.18501284875664232841136e-51 * lat ** 24 - \
        3.14445926697498412770696e-46 * lat ** 22 + \
        1.33970531285992980064454e-41 * lat ** 20 - \
        3.64729257084431972115219e-38 * lat ** 18 - \
        3.19397401776292877914625e-32 * lat ** 16 + \
        2.09755935705627800528960e-27 * lat ** 14 - \
        3.93234660296994989005722e-23 * lat ** 12 - \
        4.67346944715944948938249e-18 * lat ** 10 + \
        5.23053397296821359502194e-13 * lat ** 8 - \
        2.61499198791487361911660e-08 * lat ** 6 + \
        6.57016735969856188450312e-04 * lat ** 4 - \
        6.50322744547926518806010e+00 * lat ** 2 + \
        6378137.0


def _arc_lat_wgs84(lat: SupportedTypes,):
    r"""The arc length calculation for the wgs84 ellipsoid at a latitude uses a
    taylor series to obtain the value of the elliptic integral.

    .. math::
      arclat = \int_{0}^{lat}\sqrt{a^2 \cdot cos(lat)^2+b^2 \cdot sin(lat)^2}\ dlat

    The integral of the radius taylor series gives an arc length taylor series
    This returns the distance from the equator to a given latitude
    This is accurate to within floating point error.

    Note
    ----
    This needs to be a taylor series to avoid the elliptic integral

    Parameters
    ----------
    lat : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        2-dimensional dataset, of orthographic latitude coordinates

    Returns
    -------
    gradients : list of :class:`numpy.ndarray`, list of :class:`xarray.DataArray`
        latitudinal arc from equator calculated using th WGS84 geoid.

    See Also
    --------
    Related NCL Functions:
    `grad_latlon_cfd <https://www.ncl.ucar.edu/Document/Functions/Contributed/grad_latlon_cfd.shtml>`__,
    `gradsf <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsf.shtml>`__,
    `gradsg <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsg.shtml>`__
    """

    return \
        2.87086392358719396475614e-110 * lat ** 49 - \
        4.70057315402553703681995e-106 * lat ** 47 - \
        3.53313425533365879608994e-101 * lat ** 45 + \
        3.04345577761664668327703e-96 * lat ** 43 - \
        9.67472728333544072524681e-92 * lat ** 41 - \
        1.21447793719117046293341e-87 * lat ** 39 + \
        2.90023188160517270041180e-82 * lat ** 37 - \
        1.42116269649485608032749e-77 * lat ** 35 + \
        1.61140181088334593373753e-73 * lat ** 33 + \
        2.43380700099386902295507e-68 * lat ** 31 - \
        1.86438456247248912270599e-63 * lat ** 29 + \
        5.30488110085042076590693e-59 * lat ** 27 + \
        1.52542673636737189511924e-54 * lat ** 25 - \
        2.38613771319829866779682e-49 * lat ** 23 + \
        1.11344136742221455419521e-44 * lat ** 21 - \
        3.35038232340852202916191e-41 * lat ** 19 - \
        3.27913899018323295027126e-35 * lat ** 17 + \
        2.44062113577649370926265e-30 * lat ** 15 - \
        5.27941504241845042397655e-26 * lat ** 13 - \
        7.41522084948104995730556e-21 * lat ** 11 + \
        1.01433377184128234589268e-15 * lat ** 9 - \
        6.52003144319804471773524e-11 * lat ** 7 + \
        2.29342105667605006614878e-06 * lat ** 5 - \
        3.78342436432244021373642e-02 * lat ** 3 + \
        111319.490793273572647713 * lat


def _arc_lon_wgs84(
    lon: SupportedTypes,
    lat: SupportedTypes,
):
    r"""The arc length calculation for the wgs84 ellipsoid at a longitude uses a
    taylor series from.

    .. math::
        arclon = lon \cdot \cos(lat) \cdot \sqrt{a^2 \cdot \cos(lat)^2+b^2
        \cdot \sin(lat)^2}

    This returns the distance from the Greenwich Meridian to a given latitude
    This is accurate to within floating point error.

    Note
    ----
    This doesn't need to be a taylor series, though the taylor series
    is faster and a needed step for the _arc_lat_wgs84 function to avoid the
    elliptic integral

    Parameters
    ----------
    lon : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        2-dimensional dataset, of orthographic longitude coordinates

    lat : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        2-dimensional dataset, of orthographic latitude coordinates

    Returns
    -------
    gradients : list of :class:`numpy.ndarray`, list of :class:`xarray.DataArray`
        Longitudinal arc from Prime Meridian calculated using th WGS84 geoid.

    See Also
    --------
    Related NCL Functions:
    `grad_latlon_cfd <https://www.ncl.ucar.edu/Document/Functions/Contributed/grad_latlon_cfd.shtml>`__,
    `gradsf <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsf.shtml>`__,
    `gradsg <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsg.shtml>`__
    """

    return _rad_lat_wgs84(lat) * np.cos(lat * d2r) * lon * d2r


def gradient(data: xr.DataArray) -> [xr.DataArray]:
    r"""Extract and return the gradient values of a dataset at each point in the
    dataset. Assuming that the data points are on the surface of the WGS84
    ellipsoid.

    Parameters
    ----------
    data : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        n-dimensional dataset, with orthographic latitude longitude coordinates

    Returns
    -------
    gradients : list of :class:`numpy.ndarray`, list of :class:`xarray.DataArray`
        longitudinal and latitudinal gradients calculated using th WGS84 geoid.

    See Also
    --------
    Related NCL Functions:
    `grad_latlon_cfd <https://www.ncl.ucar.edu/Document/Functions/Contributed/grad_latlon_cfd.shtml>`__,
    `gradsf <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsf.shtml>`__,
    `gradsg <https://www.ncl.ucar.edu/Document/Functions/Built-in/gradsg.shtml>`__
    """

    lon2d, lat2d = np.meshgrid(data.coords['lon'], data.coords['lat'])

    axis0loc = xr.DataArray(
        _arc_lat_wgs84(lat2d),
        coords=data.coords,
        dims=data.dims,
    )
    axis1loc = xr.DataArray(
        _arc_lon_wgs84(lon2d, lat2d),
        coords=data.coords,
        dims=data.dims,
    )

    axis0dist = xr.DataArray(
        np.gradient(axis0loc, axis=0),
        coords=data.coords,
        dims=data.dims,
    )
    axis1dist = xr.DataArray(
        np.gradient(axis1loc, axis=1),
        coords=data.coords,
        dims=data.dims,
    )

    grad = xr.DataArray(
        np.gradient(data),
        dims=('dir',) + data.dims,
    )
    axis0grad = grad[0] / axis0dist
    axis1grad = grad[1] / axis1dist

    return [axis0grad, axis1grad]



import typing


def _generate_wrapper_docstring(wrapper_fcn: typing.Callable,
                                base_fcn: typing.Callable) -> None:
    """Generate the docstring for a wrapper function in the form of: 'This
    method is a wrapper for base_fcn', with a generated link to the base
    function.

    Parameters
    ----------
    wrapper_fcn : function
        The wrapper function to generate and assign a docstring

    base_fcn : function
        The wrapped function that the wrapper function's docstring is based
        on and links to
    """

    # create wrapper docstring
    wrapper_docstring = f".. attention:: This method is a wrapper for " \
                        f":func:`{base_fcn.__name__} <{base_fcn.__module__}.{base_fcn.__name__}>`" \
                        f"\n\n    {base_fcn.__doc__}"

    # assign docstring to wrapper function
    setattr(wrapper_fcn, '__doc__', wrapper_docstring)



import dask.array as da
import numpy as np
import typing
import warnings
import xarray as xr

from .gc_util import _generate_wrapper_docstring


def _dewtemp(
    tk: typing.Union[np.ndarray, xr.DataArray, list,
                     float], rh: typing.Union[np.ndarray, xr.DataArray, list,
                                              float]
) -> typing.Union[np.ndarray, xr.DataArray, list, float]:
    """This function calculates the dew point temperature given temperature and
    relative humidity using equations from John Dutton's "Ceaseless Wind" (pp
    273-274)

    Parameters
    ----------
    tk : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Temperature in Kelvin

    rh : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Relative humidity. Must be the same dimensions as ``temperature``

    Returns
    -------
    tdk : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Dewpoint temperature in Kelvin. Same size as input variable temperature


    See Also
    --------
    Related GeoCAT Functions:
    :func:`dewtemp`

    Related NCL Functions:
    `dewtemp_trh <https://www.ncl.ucar.edu/Document/Functions/Built-in/dewtemp_trh.shtml>`__
    """

    gc = 461.5  # gas constant for water vapor [j/{kg-k}]
    gcx = gc / (1000 * 4.186)  # [cal/{g-k}]

    lhv = (597.3 - 0.57 * (tk - 273)) / gcx
    tdk = tk * lhv / (lhv - tk * np.log(rh * 0.01))

    return tdk


def _heat_index(temperature: np.ndarray,
                relative_humidity: typing.Union[np.ndarray, xr.DataArray, list,
                                                float],
                alternate_coeffs: bool = False) -> np.ndarray:
    """Compute the 'heat index' as calculated by the National Weather Service.

    Internal function for heat_index

    Parameters
    ----------
    temperature : ndarray
        temperature(s) in Fahrenheit

    relative_humidity : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        relative humidity as a percentage. Must be the same shape as
        ``temperature``

    alternate_coeffs : bool, optional
        flag to use alternate set of coefficients appropriate for
        temperatures from 70F to 115F and humidities between 0% and 80%

    Returns
    -------
    heatindex : ndarray
        Calculated heat index. Same shape as ``temperature``


    See Also
    --------
    Related GeoCAT Functions:
    :func:`heat_index`,
    :func:`_xheat_index`

    Related NCL Functions:
    `heat_index_nws <https://www.ncl.ucar.edu/Document/Functions/Contributed/heat_index_nws.shtml>`__
    """
    # Default coefficients for (t>=80F) and (40<gh<100)
    coeffs = [
        -42.379, 2.04901523, 10.14333127, -0.22475541, -0.00683783, -0.05481717,
        0.00122874, 0.00085282, -0.00000199
    ]
    crit = [80, 40, 100]  # [T_low [F], RH_low, RH_high]

    # optional flag coefficients for (70F<t<115F) and (0<gh<80)
    # within 3F of default coeffs
    if alternate_coeffs:
        coeffs = [
            0.363445176, 0.988622465, 4.777114035, -0.114037667, -0.000850208,
            -0.020716198, 0.000687678, 0.000274954, 0.0
        ]
        crit = [70, 0, 80]  # [T_low [F], RH_low, RH_high]

    # NWS practice
    # average Steadman and t
    heatindex = (0.5 * (temperature + 61.0 + ((temperature - 68.0) * 1.2) +
                        (relative_humidity * 0.094)) + temperature) * 0.5

    # http://ehp.niehs.nih.gov/1206273/
    heatindex = xr.where(temperature < 40, temperature, heatindex)

    # if all t values less than critical, return hi
    # otherwise perform calculation
    eqtype = 0
    if not all(temperature.ravel() < crit[0]):
        eqtype = 1

        heatindex = xr.where(heatindex > crit[0],
                             _nws_eqn(coeffs, temperature, relative_humidity),
                             heatindex)

        # adjustments
        heatindex = xr.where(
            np.logical_and(relative_humidity < 13,
                           np.logical_and(temperature > 80, temperature < 112)),
            heatindex - ((13 - relative_humidity) / 4) * np.sqrt(
                (17 - abs(temperature - 95)) / 17), heatindex)

        heatindex = xr.where(
            np.logical_and(relative_humidity > 85,
                           np.logical_and(temperature > 80, temperature < 87)),
            heatindex + ((relative_humidity - 85.0) / 10.0) *
            ((87.0 - temperature) / 5.0), heatindex)

    return heatindex


def _nws_eqn(coeffs, temp, rel_hum):
    """Helper function to compute the heat index.

    Internal function for heat_index

    Parameters
    ----------
    coeffs : ndarray
        coefficients to calculate heat index

    temp : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        temperature

    rel_hum : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        relative humidity as a percentage. Must be the same shape as
        ``temperature``

    Returns
    -------
    heatindex : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        Intermediate calculated heat index. Same shape as ``temperature``


    See Also
    --------
    Related GeoCAT Functions:
    :func:`heat_index`,
    :func:`_heat_index`

    Related NCL Functions:
    `heat_index_nws <https://www.ncl.ucar.edu/Document/Functions/Contributed/heat_index_nws.shtml>`__,
    """
    heatindex = coeffs[0] \
                + coeffs[1] * temp \
                + coeffs[2] * rel_hum \
                + coeffs[3] * temp * rel_hum \
                + coeffs[4] * temp ** 2 \
                + coeffs[5] * rel_hum ** 2 \
                + coeffs[6] * temp ** 2 * rel_hum \
                + coeffs[7] * temp * rel_hum ** 2 \
                + coeffs[8] * temp ** 2 * rel_hum ** 2

    return heatindex


def _relhum(
        t: typing.Union[np.ndarray, list, float],
        w: typing.Union[np.ndarray, xr.DataArray, list,
                        float], p: typing.Union[np.ndarray, xr.DataArray, list,
                                                float]) -> np.ndarray:
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

     "Improved Magnus' Form Approx. of Saturation Vapor pressure"
     Oleg A. Alduchov and Robert E. Eskridge
     https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml

    Parameters
    ----------
    t : ndarray, :obj:`list`, or :obj:`float`
        Temperature in Kelvin

    w : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    p : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    rh : ndarray
        Relative humidity. Will have the same dimensions as ``temperature``


    See Also
    --------
    Related GeoCAT Functions:
    :func:`relhum`
    :func:`relhum_ice`
    :func:`relhum_water`
    :func:`_xrelhum`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    table = np.asarray([
        0.01403, 0.01719, 0.02101, 0.02561, 0.03117, 0.03784, 0.04584, 0.05542,
        0.06685, 0.08049, 0.09672, 0.1160, 0.1388, 0.1658, 0.1977, 0.2353,
        0.2796, 0.3316, 0.3925, 0.4638, 0.5472, 0.6444, 0.7577, 0.8894, 1.042,
        1.220, 1.425, 1.662, 1.936, 2.252, 2.615, 3.032, 3.511, 4.060, 4.688,
        5.406, 6.225, 7.159, 8.223, 9.432, 10.80, 12.36, 14.13, 16.12, 18.38,
        20.92, 23.80, 27.03, 30.67, 34.76, 39.35, 44.49, 50.26, 56.71, 63.93,
        71.98, 80.97, 90.98, 102.1, 114.5, 128.3, 143.6, 160.6, 179.4, 200.2,
        223.3, 248.8, 276.9, 307.9, 342.1, 379.8, 421.3, 466.9, 517.0, 572.0,
        632.3, 698.5, 770.9, 850.2, 937.0, 1032.0, 1146.6, 1272.0, 1408.1,
        1556.7, 1716.9, 1890.3, 2077.6, 2279.6, 2496.7, 2729.8, 2980.0, 3247.8,
        3534.1, 3839.8, 4164.8, 4510.5, 4876.9, 5265.1, 5675.2, 6107.8, 6566.2,
        7054.7, 7575.3, 8129.4, 8719.2, 9346.50, 10013.0, 10722.0, 11474.0,
        12272.0, 13119.0, 14017.0, 14969.0, 15977.0, 17044.0, 18173.0, 19367.0,
        20630.0, 21964.0, 23373.0, 24861.0, 26430.0, 28086.0, 29831.0, 31671.0,
        33608.0, 35649.0, 37796.0, 40055.0, 42430.0, 44927.0, 47551.0, 50307.0,
        53200.0, 56236.0, 59422.0, 62762.0, 66264.0, 69934.0, 73777.0, 77802.0,
        82015.0, 86423.0, 91034.0, 95855.0, 100890.0, 106160.0, 111660.0,
        117400.0, 123400.0, 129650.0, 136170.0, 142980.0, 150070.0, 157460.0,
        165160.0, 173180.0, 181530.0, 190220.0, 199260.0, 208670.0, 218450.0,
        228610.0, 239180.0, 250160.0, 261560.0, 273400.0, 285700.0, 298450.0,
        311690.0, 325420.0, 339650.0, 354410.0, 369710.0, 385560.0, 401980.0,
        418980.0, 436590.0, 454810.0, 473670.0, 493170.0, 513350.0, 534220.0,
        555800.0, 578090.0, 601130.0, 624940.0, 649530.0, 674920.0, 701130.0,
        728190.0, 756110.0, 784920.0, 814630.0, 845280.0, 876880.0, 909450.0,
        943020.0, 977610.0, 1013250.0, 1049940.0, 1087740.0, 1087740.
    ])

    maxtemp = 375.16
    mintemp = 173.16

    # replace values of t above and below max and min values for temperature
    t = np.clip(t, mintemp, maxtemp)

    it = (t - mintemp).astype(int)
    t2 = mintemp + it

    es = (t2 + 1 - t) * table[it] + (t - t2) * table[it + 1]
    es = es * 0.1

    rh = (w * (p - 0.378 * es) / (0.622 * es)) * 100

    # if any value is below 0.0001, set to 0.0001
    rh = np.clip(rh, 0.0001, None)

    return rh


def _relhum_ice(t: typing.Union[np.ndarray, list, float],
                w: typing.Union[np.ndarray, list, float],
                p: typing.Union[np.ndarray, list, float]) -> np.ndarray:
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

    "Improved Magnus' Form Approx. of Saturation Vapor pressure"
    Oleg A. Alduchov and Robert E. Eskridge
    https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml

    Parameters
    ----------
    t : ndarray, :obj:`list`, :obj:`float`
        Temperature in Kelvin

    w : ndarray, :obj:`list`, :obj:`float`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    p : ndarray, :obj:`list`, :obj:`float`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    rh : ndarray
        Relative humidity. Will have the same dimensions as ``temperature``


    See Also
    --------
    Related GeoCAT Functions:
    :func:`relhum`
    :func:`relhum_ice`
    :func:`relhum_water`
    :func:`_xrelhum`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    # Define data variables

    t0 = 273.15
    ep = 0.622
    onemep = 0.378
    es0 = 6.1128
    a = 22.571
    b = 273.71

    est = es0 * np.exp((a * (t - t0)) / ((t - t0) + b))
    qst = (ep * est) / ((p * 0.01) - onemep * est)

    rh = 100 * (w / qst)

    return rh


def _relhum_water(t: typing.Union[np.ndarray, list, float],
                  w: typing.Union[np.ndarray, list, float],
                  p: typing.Union[np.ndarray, list, float]) -> np.ndarray:
    """Calculates relative humidity with respect to water, given temperature,
    mixing ratio, and pressure.

    Definition of mixing ratio if:

    - ``es``  - is the saturation mixing ratio
    - ``ep``  - is the ratio of the molecular weights of water vapor to dry air
    - ``p``   - is the atmospheric pressure
    - ``rh``  - is the relative humidity (given as a percent)

    .. math::
        rh =  100*  q / ( (ep*es)/(p-es) )

    Parameters
    ----------
    t : ndarray, :obj:`list`, :obj:`float`
        Temperature in Kelvin

    w : ndarray, :obj:`list`, :obj:`float`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    p : ndarray, :obj:`list`, :obj:`float`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    rh : ndarray
        Relative humidity. Will have the same dimensions as ``temperature``

    See Also
    --------
    Related GeoCAT Functions:
    :func:`relhum`
    :func:`relhum_ice`
    :func:`relhum_water`
    :func:`_xrelhum`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    # Define data variables

    t0 = 273.15
    ep = 0.622
    onemep = 0.378
    es0 = 6.1128
    a = 17.269
    b = 35.86

    est = es0 * np.exp((a * (t - t0)) / (t - b))
    qst = (ep * est) / ((p * 0.01) - onemep * est)

    rh = 100 * (w / qst)

    return rh


def _xheat_index(temperature: xr.DataArray,
                 relative_humidity: xr.DataArray,
                 alternate_coeffs: bool = False) -> tuple([xr.DataArray, int]):
    """Compute the 'heat index' as calculated by the National Weather Service.

    Internal function for heat_index for dask

    Parameters
    ----------
    temperature : :class:`xarray.DataArray`
        temperature(s) in Fahrenheit

    relative_humidity : :class:`xarray.DataArray`
        relative humidity as a percentage. Must be the same shape as
        ``temperature``

    alternate_coeffs : bool, optional
        flag to use alternate set of coefficients appropriate for
        temperatures from 70F to 115F and humidities between 0% and 80%

    Returns
    -------
    heatindex : :class:`xarray.DataArray`
        Calculated heat index. Same shape as ``temperature``

    eqtype : :class:`int`
        version of equations used, for xarray attrs output

    See Also
    --------
    Related GeoCAT Functions:
    :func:`heat_index`
    :func:`_heat_index`

    Related NCL Functions:
    `heat_index_nws <https://www.ncl.ucar.edu/Document/Functions/Contributed/heat_index_nws.shtml>`__,
    """
    # Default coefficients for (t>=80F) and (40<gh<100)
    coeffs = [
        -42.379, 2.04901523, 10.14333127, -0.22475541, -0.00683783, -0.05481717,
        0.00122874, 0.00085282, -0.00000199
    ]
    crit = [80, 40, 100]  # [T_low [F], RH_low, RH_high]

    # optional flag coefficients for (70F<t<115F) and (0<gh<80)
    # within 3F of default coeffs
    if alternate_coeffs:
        coeffs = [
            0.363445176, 0.988622465, 4.777114035, -0.114037667, -0.000850208,
            -0.020716198, 0.000687678, 0.000274954, 0.0
        ]
        crit = [70, 0, 80]  # [T_low [F], RH_low, RH_high]

    # NWS practice
    # average Steadman and t
    heatindex = (0.5 * (temperature + 61.0 + ((temperature - 68.0) * 1.2) +
                        (relative_humidity * 0.094)) + temperature) * 0.5

    # http://ehp.niehs.nih.gov/1206273/
    heatindex = xr.where(temperature < 40, temperature, heatindex)

    # if all t values less than critical, return hi
    # otherwise perform calculation
    eqtype = 0
    if not all(temperature.data.ravel() < crit[0]):
        eqtype = 1

        heatindex = xr.where(heatindex > crit[0],
                             _nws_eqn(coeffs, temperature, relative_humidity),
                             heatindex)

        # adjustments
        heatindex = xr.where(
            np.logical_and(relative_humidity < 13,
                           np.logical_and(temperature > 80, temperature < 112)),
            heatindex - ((13 - relative_humidity) / 4) * np.sqrt(
                (17 - abs(temperature - 95)) / 17), heatindex)

        heatindex = xr.where(
            np.logical_and(relative_humidity > 85,
                           np.logical_and(temperature > 80, temperature < 87)),
            heatindex + ((relative_humidity - 85.0) / 10.0) *
            ((87.0 - temperature) / 5.0), heatindex)

    return heatindex, eqtype


def _xrelhum(t: xr.DataArray, w: xr.DataArray, p: xr.DataArray) -> xr.DataArray:
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

     "Improved Magnus' Form Approx. of Saturation Vapor pressure"
     Oleg A. Alduchov and Robert E. Eskridge
     https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml

    Parameters
    ----------
    t : :class:`xarray.DataArray`
        Temperature in Kelvin

    w : :class:`xarray.DataArray`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    p : :class:`xarray.DataArray`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    rh : :class:`xarray.DataArray`
        Relative humidity. Will have the same dimensions as ``temperature``


    See Also
    --------
    Related GeoCAT Functions:
    :func:`relhum`
    :func:`relhum_ice`
    :func:`relhum_water`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    table = da.from_array([
        0.01403, 0.01719, 0.02101, 0.02561, 0.03117, 0.03784, 0.04584, 0.05542,
        0.06685, 0.08049, 0.09672, 0.1160, 0.1388, 0.1658, 0.1977, 0.2353,
        0.2796, 0.3316, 0.3925, 0.4638, 0.5472, 0.6444, 0.7577, 0.8894, 1.042,
        1.220, 1.425, 1.662, 1.936, 2.252, 2.615, 3.032, 3.511, 4.060, 4.688,
        5.406, 6.225, 7.159, 8.223, 9.432, 10.80, 12.36, 14.13, 16.12, 18.38,
        20.92, 23.80, 27.03, 30.67, 34.76, 39.35, 44.49, 50.26, 56.71, 63.93,
        71.98, 80.97, 90.98, 102.1, 114.5, 128.3, 143.6, 160.6, 179.4, 200.2,
        223.3, 248.8, 276.9, 307.9, 342.1, 379.8, 421.3, 466.9, 517.0, 572.0,
        632.3, 698.5, 770.9, 850.2, 937.0, 1032.0, 1146.6, 1272.0, 1408.1,
        1556.7, 1716.9, 1890.3, 2077.6, 2279.6, 2496.7, 2729.8, 2980.0, 3247.8,
        3534.1, 3839.8, 4164.8, 4510.5, 4876.9, 5265.1, 5675.2, 6107.8, 6566.2,
        7054.7, 7575.3, 8129.4, 8719.2, 9346.50, 10013.0, 10722.0, 11474.0,
        12272.0, 13119.0, 14017.0, 14969.0, 15977.0, 17044.0, 18173.0, 19367.0,
        20630.0, 21964.0, 23373.0, 24861.0, 26430.0, 28086.0, 29831.0, 31671.0,
        33608.0, 35649.0, 37796.0, 40055.0, 42430.0, 44927.0, 47551.0, 50307.0,
        53200.0, 56236.0, 59422.0, 62762.0, 66264.0, 69934.0, 73777.0, 77802.0,
        82015.0, 86423.0, 91034.0, 95855.0, 100890.0, 106160.0, 111660.0,
        117400.0, 123400.0, 129650.0, 136170.0, 142980.0, 150070.0, 157460.0,
        165160.0, 173180.0, 181530.0, 190220.0, 199260.0, 208670.0, 218450.0,
        228610.0, 239180.0, 250160.0, 261560.0, 273400.0, 285700.0, 298450.0,
        311690.0, 325420.0, 339650.0, 354410.0, 369710.0, 385560.0, 401980.0,
        418980.0, 436590.0, 454810.0, 473670.0, 493170.0, 513350.0, 534220.0,
        555800.0, 578090.0, 601130.0, 624940.0, 649530.0, 674920.0, 701130.0,
        728190.0, 756110.0, 784920.0, 814630.0, 845280.0, 876880.0, 909450.0,
        943020.0, 977610.0, 1013250.0, 1049940.0, 1087740.0, 1087740.
    ])

    maxtemp = 375.16
    mintemp = 173.16

    # replace values of t above and below max and min values for temperature
    t = t.clip(mintemp, maxtemp)

    it = (t - mintemp).astype(int).data
    t2 = mintemp + it

    it_shape = it.shape

    es = (t2 + 1 - t) * table[it.ravel()].reshape(it_shape) + (
        t - t2) * table[it.ravel() + 1].reshape(it_shape)
    es = es * 0.1

    rh = (w * (p - 0.378 * es) / (0.622 * es)) * 100

    rh = rh.clip(0.0001, None)

    return rh


def dewtemp(
    temperature: typing.Union[np.ndarray, xr.DataArray, list, float],
    relative_humidity: typing.Union[np.ndarray, xr.DataArray, list, float]
) -> typing.Union[np.ndarray, float]:
    """This function calculates the dew point temperature given temperature and
    relative humidity using equations from John Dutton's "Ceaseless Wind" (pp
    273-274)

    Parameters
    ----------
    temperature : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Temperature in Kelvin

    relative_humidity : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Relative humidity. Must be the same dimensions as ``temperature``

    Returns
    -------
    dew_pnt_temp : ndarray or :obj:`float`
        Dewpoint temperature in Kelvin. Same size as input variable temperature

    See Also
    --------
    Related GeoCAT Functions:
    :func:`_dewtemp`

    Related NCL Functions:
    `dewtemp_trh <https://www.ncl.ucar.edu/Document/Functions/Built-in/dewtemp_trh.shtml>`__
    """

    inputs = [temperature, relative_humidity]

    # ensure all inputs same size
    if not (np.shape(x) == np.shape(inputs[0]) for x in inputs):
        raise ValueError("dewtemp: dimensions of inputs are not the same")

    # Get input types
    in_types = [type(item) for item in inputs]

    if xr.DataArray in in_types:

        # check all inputs are xarray.DataArray
        if any(x != xr.DataArray for x in in_types):
            raise TypeError(
                "relhum: if using xarray, all inputs must be xarray")

        # call internal computation function
        # note: no alternative internal function required for dewtemp
        dew_pnt_temp = _dewtemp(temperature, relative_humidity)

        # set xarray attributes
        dew_pnt_temp.attrs['long_name'] = 'dew point temperature'
        dew_pnt_temp.attrs['units'] = 'Kelvin'

    else:
        # ensure in numpy array for function call
        temperature = np.asarray(temperature)
        relative_humidity = np.asarray(relative_humidity)

        # function call for non-dask/xarray
        dew_pnt_temp = _dewtemp(temperature, relative_humidity)

    return dew_pnt_temp


def heat_index(
        temperature: typing.Union[np.ndarray, xr.DataArray, list, float],
        relative_humidity: typing.Union[np.ndarray, xr.DataArray, list, float],
        alternate_coeffs: bool = False
) -> typing.Union[np.ndarray, xr.DataArray]:
    """Compute the 'heat index' as calculated by the National Weather Service.

    The heat index calculation in this funtion is described at:
    https://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml

    The 'Heat Index' is a measure of how hot weather "feels" to the body. The combination of temperature and humidity
    produce an "apparent temperature" or the temperature the body "feels". The returned values are for shady
    locations only. Exposure to full sunshine can increase heat index values by up to 15°F. Also, strong winds,
    particularly with very hot, dry air, can be extremely hazardous as the wind adds heat to the body

    The computation of the heat index is a refinement of a result obtained by multiple regression analysis carried
    out by Lans P. Rothfusz and described in a 1990 National Weather Service (NWS) Technical Attachment (SR 90-23).
    All values less that 40F/4.4C/277.65K are set to the ambient temperature.

    In practice, the Steadman formula is computed first and the result averaged with the temperature. If this heat
    index value is 80 degrees F or higher, the full regression equation along with any adjustment as described above
    is applied. If the ambient temperature is less the 40F (4.4C/277.65K), the heat index is set to the ambient
    temperature.

    Parameters
    ----------
    temperature : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        temperature(s) in Fahrenheit

    relative_humidity : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        relative humidity as a percentage. Must be the same shape as
        ``temperature``

    alternate_coeffs : bool, optional
        flag to use alternate set of coefficients appropriate for
        temperatures from 70F to 115F and humidities between 0% and 80%

    Returns
    -------
    heatindex : ndarray, :class:`xarray.DataArray`
        Calculated heat index. Same shape as ``temperature``

    Examples
    --------
    >>> import numpy as np
    >>> import geocat.comp
    >>> t = np.array([104, 100, 92])
    >>> rh = np.array([55, 65, 60])
    >>> hi = heat_index(t,rh)
    >>> hi
    array([137.36135724, 135.8679973 , 104.68441864])


    See Also
    --------
    Related GeoCAT Functions:
    :func:`_heat_index`
    :func:`_xheat_index`

    Related NCL Functions:
    `heat_index_nws <https://www.ncl.ucar.edu/Document/Functions/Contributed/heat_index_nws.shtml>`__
    """

    inputs = [temperature, relative_humidity]

    # ensure all inputs same size
    if not (np.shape(x) == np.shape(inputs[0]) for x in inputs):
        raise ValueError("heat_index: dimensions of inputs are not the same")

    # Get input types
    in_types = [type(item) for item in inputs]

    # run dask compatible version if input is xarray
    if xr.DataArray in in_types:

        # check all inputs are xarray.DataArray
        if not all(x == xr.DataArray for x in in_types):
            raise TypeError(
                "heat_index: if using xarray, all inputs must be xarray")

        # input validation on relative humidity
        if any(relative_humidity.data.ravel() < 0) or any(
                relative_humidity.data.ravel() > 100):
            raise ValueError('heat_index: invalid values for relative humidity')

        # Check if relative humidity fractional
        if all(relative_humidity.data.ravel() < 1):
            warnings.warn(
                "WARNING: rh must be %, not fractional; All rh are < 1")

        # call internal computation function
        heatindex, eqtype = _xheat_index(temperature, relative_humidity,
                                         alternate_coeffs)

        # set xarray attributes
        heatindex.attrs['long_name'] = "heat index: NWS"
        heatindex.attrs['units'] = "F"
        heatindex.attrs[
            'www'] = "https://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml"
        heatindex.attrs['info'] = "appropriate for shady locations with no wind"

        if eqtype == 1:
            heatindex.attrs[
                'tag'] = "NCL: heat_index_nws; (Steadman+t)*0.5 and Rothfusz"
        else:
            heatindex.attrs['tag'] = "NCL: heat_index_nws; (Steadman+t)*0.5"

    else:
        # ensure in numpy array for function call
        temperature = np.atleast_1d(temperature)
        relative_humidity = np.atleast_1d(relative_humidity)

        # input validation on relative humidity
        if any(relative_humidity.ravel() < 0) or any(
                relative_humidity.ravel() > 100):
            raise ValueError('heat_index: invalid values for relative humidity')

        # Check if relative humidity fractional
        if all(relative_humidity.ravel() < 1):
            warnings.warn(
                "WARNING: rh must be %, not fractional; All rh are < 1")

        # function call for non-dask/xarray
        heatindex = _heat_index(temperature, relative_humidity,
                                alternate_coeffs)

    return heatindex


def relhum(
    temperature: typing.Union[np.ndarray, xr.DataArray, list, float],
    mixing_ratio: typing.Union[np.ndarray, xr.DataArray, list, float],
    pressure: typing.Union[np.ndarray, xr.DataArray, list, float]
) -> typing.Union[np.ndarray, xr.DataArray]:
    """This function calculates the relative humidity given temperature, mixing
    ratio, and pressure.

    "Improved Magnus' Form Approx. of Saturation Vapor pressure"
    Oleg A. Alduchov and Robert E. Eskridge
    https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml

    Parameters
    ----------
    temperature : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Temperature in Kelvin

    mixing_ratio : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    pressure : ndarray, :class:`xarray.DataArray`, :obj:`list`, or :obj:`float`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    relative_humidity : ndarray or :class:`xarray.DataArray`
        Relative humidity. Will have the same dimensions as ``temperature``


    See Also
    --------
    Related GeoCAT Functions:
    :func:`_xrelhum`
    :func:`relhum_ice`
    :func:`relhum_water`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    inputs = [temperature, mixing_ratio, pressure]

    # ensure all inputs same size
    if not (np.shape(x) == np.shape(inputs[0]) for x in inputs):
        raise ValueError("relhum: dimensions of inputs are not the same")

    # Get input types
    in_types = [type(item) for item in inputs]

    # run dask compatible version if input is xarray
    if xr.DataArray in in_types:

        # check all inputs are xarray.DataArray
        if not all(x == xr.DataArray for x in in_types):
            raise TypeError(
                "relhum: if using xarray, all inputs must be xarray")

        # call internal computation function
        relative_humidity = _xrelhum(temperature, mixing_ratio, pressure)

        # set xarray attributes
        relative_humidity.attrs['long_name'] = "relative humidity"
        relative_humidity.attrs['units'] = 'percentage'
        relative_humidity.attrs[
            'info'] = 'https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml'

    else:
        # ensure in numpy array for function call
        temperature = np.asarray(temperature)
        mixing_ratio = np.asarray(mixing_ratio)
        pressure = np.asarray(pressure)

        # function call for non-dask/xarray
        relative_humidity = _relhum(temperature, mixing_ratio, pressure)

    return relative_humidity


def relhum_ice(temperature: typing.Union[np.ndarray, list, float],
               mixing_ratio: typing.Union[np.ndarray, list, float],
               pressure: typing.Union[np.ndarray, list, float]) -> np.ndarray:
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

    "Improved Magnus' Form Approx. of Saturation Vapor pressure"
    Oleg A. Alduchov and Robert E. Eskridge
    https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml

    Parameters
    ----------
    temperature : ndarray, :obj:`list`, or :obj:`float`
        Temperature in Kelvin

    mixing_ratio : ndarray, :obj:`list`, or :obj:`float`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    pressure : ndarray, :obj:`list`, or :obj:`float`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    relative_humidity : ndarray
        Relative humidity. Will have the same dimensions as ``temperature``

    See Also
    --------
    Related GeoCAT Functions:
    :func:`relhum`
    :func:`_xrelhum`
    :func:`relhum_water`
    :func:`_relhum_ice`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    # If xarray input, pull data and store metadata
    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords
        save_attrs = temperature.attrs

    # ensure in numpy array for function call
    temperature = np.asarray(temperature)
    mixing_ratio = np.asarray(mixing_ratio)
    pressure = np.asarray(pressure)

    # ensure all inputs same size
    if np.shape(temperature) != np.shape(mixing_ratio) or np.shape(
            temperature) != np.shape(pressure):
        raise ValueError(f"relhum_ice: dimensions of inputs are not the same")

    relative_humidity = _relhum_ice(temperature, mixing_ratio, pressure)

    # output as xarray if input as xarray
    if x_out:
        relative_humidity = xr.DataArray(data=relative_humidity,
                                         coords=save_coords,
                                         dims=save_dims,
                                         attrs=save_attrs)

    return relative_humidity


def relhum_water(temperature: typing.Union[np.ndarray, list, float],
                 mixing_ratio: typing.Union[np.ndarray, list, float],
                 pressure: typing.Union[np.ndarray, list, float]) -> np.ndarray:
    """Calculates relative humidity with respect to water, given temperature,
    mixing ratio, and pressure.

    Definition of mixing ratio if:

    - `es`  - is the saturation mixing ratio
    - `ep`  - is the ratio of the molecular weights of water vapor to dry air
    - `p`   - is the atmospheric pressure
    - `rh`  - is the relative humidity (given as a percent)

    .. math::
        rh =  100  q / ( (ep*es)/(p-es) )

    Parameters
    ----------
    temperature : ndarray, :obj:`list`, or :obj:`float`
        Temperature in Kelvin

    mixing_ratio : ndarray, :obj:`list`, or :obj:`float`
        Mixing ratio in kg/kg. Must have the same dimensions as ``temperature``

    pressure : ndarray, :obj:`list`, or :obj:`float`
        Pressure in Pa. Must have the same dimensions as ``temperature``

    Returns
    -------
    relative_humidity : ndarray
        Relative humidity. Will have the same dimensions as ``temperature``

    See Also
    --------
    Related GeoCAT Functions:
    :func:`relhum`
    :func:`_xrelhum`
    :func:`relhum_ice`
    :func:`relhum_water`

    Related NCL Functions:
    `relhum <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum.shtml>`__,
    `relhum_ttd <https://www.ncl.ucar.edu/Document/Functions/Contributed/relhum_ttd.shtml>`__,
    `relhum_ice <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_ice.shtml>`__,
    `relhum_water <https://www.ncl.ucar.edu/Document/Functions/Built-in/relhum_water.shtml>`__
    """

    # If xarray input, pull data and store metadata
    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords
        save_attrs = temperature.attrs

    # ensure in numpy array for function call
    temperature = np.asarray(temperature)
    mixing_ratio = np.asarray(mixing_ratio)
    pressure = np.asarray(pressure)

    # ensure all inputs same size
    if np.shape(temperature) != np.shape(mixing_ratio) or np.shape(
            temperature) != np.shape(pressure):
        raise ValueError(f"relhum_water: dimensions of inputs are not the same")

    relative_humidity = _relhum_water(temperature, mixing_ratio, pressure)

    # output as xarray if input as xarray
    if x_out:
        relative_humidity = xr.DataArray(data=relative_humidity,
                                         coords=save_coords,
                                         dims=save_dims,
                                         attrs=save_attrs)

    return relative_humidity


def max_daylight(
    jday: typing.Union[np.ndarray, xr.DataArray, list,
                       float], lat: typing.Union[np.ndarray, xr.DataArray, list,
                                                 float]
) -> typing.Union[np.ndarray, xr.DataArray, float]:
    """Computes maximum number of daylight hours as described in the Food and
    Agriculture Organization (FAO) Irrigation and Drainage Paper 56 entitled:

    Crop evapotranspiration - Guidelines for computing crop water
    requirement. Specifically, see equation 34 of Chapter 3.

    Note for abs(lat) > 55 the eqns have limited validity.

    Parameters
    ----------
    jday : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        Day of year. Must be 1D

    lat : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        Latitude in degrees. Must be 1D

    Returns
    -------
    sunmax : ndarray, :class:`xarray.DataArray`, :class:`float`
        Calculated maximum sunlight in hours/day

    Examples
    --------
    >>> from geocat.comp import max_daylight
    >>> import numpy as np
    >>> jday = np.array([100, 123, 246])
    >>> lat = np.array([10, 20])
    >>> max_daylight(jday, lat)
    array([[12.18035083, 12.37238906],
           [12.37577081, 12.77668231],
           [12.16196585, 12.33440805]])


    See Also
    --------
    Related NCL Functions:
    `daylight_fao56 <https://www.ncl.ucar.edu/Document/Functions/Crop/daylight_fao56.shtml>`__
    """

    x_out = False
    if isinstance(jday, xr.DataArray):
        x_out = True

    # convert inputs to numpy arrays for function call if necessary
    if not xr.core.utils.is_duck_array(jday):
        jday = np.asarray(jday, dtype='float32')
    if not xr.core.utils.is_duck_array(lat):
        lat = np.asarray(lat, dtype='float32')

    # check to ensure dimension of lat is not greater than two
    if lat.ndim > 1 or jday.ndim > 1:
        raise ValueError('max_daylight: inputs must have at most one dimension')

    # check if latitude is outside of acceptable ranges
    # warn if more than abs(55)
    # Give stronger warning if more than abs(66)
    if (abs(lat) > 55).all() and (abs(lat) <= 66).all():
        warnings.warn(
            "WARNING: max_daylight has limited validity for abs(lat) > 55 ")
    elif (abs(lat) > 66).all():
        warnings.warn(
            'WARNING: max_daylight: calculation not possible for abs(lat) > 66 for all values of jday, '
            'errors may occur')

    # define constants
    pi = np.pi
    rad = pi / 180
    pi2yr = 2 * pi / 365
    latrad = lat * rad
    con = 24 / pi

    # Equation 24 from FAO56
    sdec = 0.409 * np.sin(pi2yr * jday - 1.39)

    # Equation 25 from FAO56
    ws = np.arccos(np.outer(-np.tan(latrad), np.tan(sdec)))

    # Equation 34 from FAO56
    dlm = np.transpose(con * ws)

    # handle metadata if xarray output
    if x_out:
        dlm = xr.DataArray(dlm, coords=[jday, lat], dims=["doy", "lat"])
        dlm.attrs['long_name'] = "maximum daylight: FAO_56"
        dlm.attrs['units'] = "hours/day"
        dlm.attrs['url'] = "http://www.fao.org/docrep/X0490E/x0490e07.htm"
        dlm.attrs['info'] = "FAO 56; EQN 34; max_daylight"

    return dlm


def psychrometric_constant(
    pressure: typing.Union[np.ndarray, xr.DataArray, list, float]
) -> typing.Union[np.ndarray, xr.DataArray]:
    """Compute psychrometric constant [kPa / C] as described in the Food and
    Agriculture Organization (FAO) Irrigation and Drainage Paper 56 entitled:

    Crop evapotranspiration - Guidelines for computing crop water
    requirement. Specifically, see equation 7 of Chapter 3 or equation 3-2 in
    Annex 3.

    From FAO 56:

    The specific heat at constant pressure is the amount of energy required to
    increase the temperature of a unit mass of air by one degree at constant
    pressure. Its value depends on the composition of the air, i.e.,  on its
    humidity. For average atmospheric conditions a value
    cp = 1.013 10-3 MJ kg-1 C-1 can be used. As an average atmospheric
    pressure is used for each location (Equation 7), the psychrometric
    constant is kept constant for each location.

    A table listing the psychrometric constant for different altitudes is
    located here: https://www.fao.org/3/X0490E/x0490e0j.htm

    Parameters
    ----------
    pressure : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        pressure in kPa/C

    Returns
    -------
    psy_const : ndarray, :class:`xarray.DataArray`
        the computed psychrometric constant. Same shape as pressure.

    Examples
    --------
    >>> import numpy as np
    >>> from geocat.comp import psychrometric_constant
    >>> pressure = np.array([60, 80, 100])
    >>> psychrometric_constant(pressure)
    array([0.0398844, 0.0531792, 0.066474 ])


    See Also
    --------
    Related NCL Functions:
    `psychro_fao56 <https://www.ncl.ucar.edu/Document/Functions/Crop/psychro_fao56.shtml>`__
    """

    # Constant
    con = 0.66474e-3

    in_type = type(pressure)

    # Psychrometric constant calculation
    # if input not xarray, make sure in numpy for calculation
    if in_type is not xr.DataArray:
        psy_const = con * np.asarray(pressure)

    # else if input is xarray, add relevant metadata for xarray output
    else:
        psy_const = con * pressure
        psy_const.attrs['long_name'] = "psychrometric constant"
        psy_const.attrs['units'] = "kPa/C"
        psy_const.attrs[
            'url'] = "https://www.fao.org/docrep/X0490E/x0490e07.htm"
        psy_const.attrs['info'] = "FAO 56; EQN 8; psychrometric_constant"

    return psy_const


def saturation_vapor_pressure(
    temperature: typing.Union[np.ndarray, xr.DataArray, list, float],
    tfill: typing.Union[float] = np.nan
) -> typing.Union[np.ndarray, xr.DataArray]:
    """Compute saturation vapor pressure as described in the Food and
    Agriculture Organization (FAO) Irrigation and Drainage Paper 56
    entitled:

    Crop evapotranspiration - Guidelines for computing crop water
    requirement. Specifically, see equation 11 of Chapter 3.

    This is Tetens' Formula: an empirical expression for saturation vapor
    pressure with respect to liquid water that includes the variation of
    latent heat with temperature.

    Note that if temperature = tdew, then this function computes actual vapor
    pressure.

    Parameters
    ----------
    temperature : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        Temperature in Fahrenheit

    tfill : float, numpy.nan, optional
        An optional parameter for a fill value in the return value

    Returns
    -------
    svp : ndarray, :class:`xarray.DataArray`
        the computed actual saturation vapor pressure in kPa.
        Same shape as temperature.

    Examples
    --------
    >>> import numpy as np
    >>> from geocat.comp import saturation_vapor_pressure
    >>> temp = np.array([50, 60, 70])
    >>> saturation_vapor_pressure(temp)
    array([1.22796262, 1.76730647, 2.50402976])


    See Also
    --------
    Related GeoCAT Functions:
    :func:`actual_saturation_vapor_pressure`
    :func:`saturation_vapor_pressure_slope`

    Related NCL Functions:
    `satvpr_temp_fao56 <https://www.ncl.ucar.edu/Document/Functions/Crop/satvpr_temp_fao56.shtml>`__
    """

    in_type = type(temperature)

    if in_type is xr.DataArray:

        # convert temperature to Celsius
        temp_c = (temperature - 32) * 5 / 9

        # calculate svp
        svp = xr.where(temp_c > 0, 0.6108 * np.exp(
            (17.27 * temp_c) / (temp_c + 237.3)), tfill)

        # add relevant metadata
        svp.attrs['long_name'] = "saturation vapor pressure"
        svp.attrs['units'] = "kPa"
        svp.attrs['url'] = "https://www.fao.org/docrep/X0490E/x0490e07.htm"
        svp.attrs['info'] = "FAO 56; EQN 11; saturation_vapor_pressure"

    else:
        temperature = np.asarray(temperature)

        temp_c = (temperature - 32) * 5 / 9
        svp = np.where(temp_c > 0, 0.6108 * np.exp(
            (17.27 * temp_c) / (temp_c + 237.3)), tfill)

    return svp


def actual_saturation_vapor_pressure(
    tdew: typing.Union[np.ndarray, xr.DataArray, list, float],
    tfill: typing.Union[float] = np.nan
) -> typing.Union[np.ndarray, xr.DataArray]:
    """Compute 'actual' saturation vapor pressure [kPa] as described in the
    Food and Agriculture Organization (FAO) Irrigation and Drainage Paper 56
    entitled:

    Crop evapotranspiration - Guidelines for computing crop water
    requirement. Specifically, see equation 14 of Chapter 3.

    The dew point temperature is synonymous with the wet bulb temperature.

    Note that this function is the same as saturation_vapor_pressure, but with
    temperature = dew point temperature with different metadata

    Parameters
    ----------
    tdew : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        Dew point temperatures in Fahrenheit

    tfill : float, numpy.nan, optional
        An optional parameter for a fill value in the return value

    Returns
    -------
    asvp : ndarray, :class:`xarray.DataArray`
        the computed actual saturation vapor pressure in kPa.
        Same shape as tdew.

    Examples
    --------
    >>> import numpy as np
    >>> from geocat.comp import actual_saturation_vapor_pressure
    >>> temp = np.array([50, 60, 70])
    >>> actual_saturation_vapor_pressure(temp)
    array([1.22796262, 1.76730647, 2.50402976])


    See Also
    --------
    Related GeoCAT Functions:
    :func:`saturation_vapor_pressure`
    :func:`saturation_vapor_pressure_slope`

    Related NCL Functions:
    `satvpr_tdew_fao56 <https://www.ncl.ucar.edu/Document/Functions/Crop/satvpr_tdew_fao56.shtml>`__
    """

    in_type = type(tdew)

    asvp = saturation_vapor_pressure(tdew, tfill)

    # reformat metadata for xarray
    if in_type is xr.DataArray:
        asvp.attrs['long_name'] = "actual saturation vapor pressure via Tdew"
        asvp.attrs['units'] = "kPa"
        asvp.attrs['url'] = "https://www.fao.org/docrep/X0490E/x0490e07.htm"
        asvp.attrs['info'] = "FAO 56; EQN 14; actual_saturation_vapor_pressure"

    return asvp


def saturation_vapor_pressure_slope(
    temperature: typing.Union[np.ndarray, xr.DataArray, list, float],
    tfill: typing.Union[float] = np.nan
) -> typing.Union[np.ndarray, xr.DataArray]:
    """Compute the slope [kPa/C] of saturation vapor pressure curve as
    described in the Food and Agriculture Organization (FAO) Irrigation and
    Drainage Paper 56 entitled:

    Crop evapotranspiration - Guidelines for computing crop water
    requirement. Specifically, see equation 13 of Chapter 3.

    Parameters
    ----------
    temperature : ndarray, :class:`xarray.DataArray`, :class:`list`, :class:`float`
        Temperature in Fahrenheit

    tfill : float, numpy.nan, optional
        An optional parameter for a fill value in the return value

    Returns
    -------
    svp_slope : ndarray, :class:`xarray.DataArray`
        The computed slopes of the saturation vapor pressure curve.
        Will be the same shape as temperature.

    Examples
    --------
    >>> import numpy as np
    >>> from geocat.comp import saturation_vapor_pressure_slope
    >>> temp = np.array([50, 60, 70])
    >>> saturation_vapor_pressure_slope(temp)
    array([0.08224261, 0.11322096, 0.153595  ])


    See Also
    --------
    Related GeoCAT Functions:
    :func:`actual_saturation_vapor_pressure`
    :func:`saturation_vapor_pressure_slope`

    Related NCL Functions:
    `satvpr_temp_fao56 <https://www.ncl.ucar.edu/Document/Functions/Crop/satvpr_temp_fao56.shtml>`__
    """
    in_type = type(temperature)

    if in_type is xr.DataArray:

        # convert to Celsius
        temp_c = (temperature - 32) * 5 / 9

        # calculate svp_slope
        svp_slope = xr.where(
            temp_c > 0, 4096 * (0.6108 * np.exp(
                (17.27 * temp_c) / (temp_c + 237.3)) / (temp_c + 237.3)**2),
            tfill)

        # add relevant metadata
        svp_slope.attrs['long_name'] = "slope saturation vapor pressure curve"
        svp_slope.attrs['units'] = "kPa/C"
        svp_slope.attrs[
            'url'] = "https://www.fao.org/docrep/X0490E/x0490e07.htm"
        svp_slope.attrs[
            'info'] = "FAO 56; EQN 13; saturation_vapor_pressure_slope"

    else:
        temperature = np.asarray(temperature)

        # convert to Celsius
        temp_c = (temperature - 32) * 5 / 9

        # calculate svp_slope
        svp_slope = np.where(
            temp_c > 0, 4096 * (0.6108 * np.exp(
                (17.27 * temp_c) / (temp_c + 237.3)) / (temp_c + 237.3)**2),
            tfill)

    return svp_slope


def _delta_pressure1D(pressure_lev, surface_pressure):
    """Helper function for `delta_pressure`. Calculates the pressure layer
    thickness (delta pressure) of a one-dimensional pressure level array.

    Returns an array of length matching `pressure_lev`.

    Parameters
    ----------
    pressure_lev : :class:`numpy.ndarray`
        The pressure level array. May be in ascending or descending order.
        Must have the same units as `surface_pressure`.

    surface_pressure : :class:`float`
        The scalar surface pressure. Must have the same units as
        `pressure_lev`.

    Returns
    -------
    delta_pressure : :class:`numpy.ndarray`
        The pressure layer thickness array. Shares dimensions and units of
        `pressure_lev`.
    """
    pressure_top = min(pressure_lev)

    # Safety checks
    if pressure_top <= 0:
        warnings.warn("'pressure_lev` values must all be positive.")
    if pressure_top > surface_pressure:
        warnings.warn(
            "`surface_pressure` must be greater than minimum `pressure_lev` value."
        )

    # Sort so pressure increases (array goes from top of atmosphere to bottom)
    is_pressuredecreasing = pressure_lev[1] < pressure_lev[0]
    if is_pressuredecreasing:
        pressure_lev = np.flip(pressure_lev)

    # Calculate delta pressure
    delta_pressure = np.empty_like(pressure_lev)

    delta_pressure[0] = (pressure_lev[0] +
                         pressure_lev[1]) / 2 - pressure_top  # top level
    delta_pressure[1:-1] = [
        (a - b) / 2 for a, b in zip(pressure_lev[2:], pressure_lev[:-1])
    ]
    delta_pressure[-1] = surface_pressure - (
        pressure_lev[-1] + pressure_lev[-2]) / 2  # bottom level

    # Return delta_pressure to original order
    if is_pressuredecreasing:
        delta_pressure = np.flip(delta_pressure)

    return delta_pressure


def delta_pressure(pressure_lev, surface_pressure):
    """Calculates the pressure layer thickness (delta pressure) of a constant
    pressure level coordinate system.

    Returns an array of shape matching (``surface_pressure``, ``pressure_lev``).

    Parameters
    ----------
    pressure_lev : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        The pressure level array. May be in ascending or descending order.
        Must have the same units as ``surface_pressure``.

    surface_pressure : :class:`int`, :class:`float`, :class:`numpy.ndarray`, :class:`xarray.DataArray`
        The scalar or N-dimensional surface pressure array. Must have the same
        units as ``pressure_lev``.

    Returns
    -------
    delta_pressure : :class:`numpy.ndarray`, :class:`xarray.DataArray`
        The pressure layer thickness array. Shares units with ``pressure_lev``.
        If ``surface_pressure`` is scalar, shares dimensions with
        ``pressure_level``. If ``surface_pressure`` is an array than the returned
        array will have an additional dimension [e.g. (lat, lon, time) becomes
        (lat, lon, time, lev)]. Will always be the same type as ``surface_pressure``.

    See Also
    --------
    Related NCL Functions:
    `dpres_plev <https://www.ncl.ucar.edu/Document/Functions/Built-in/dpres_plevel.shtml>`__
    """
    # Get original array types
    type_surface_pressure = type(
        surface_pressure
    )  # save type for delta_pressure to same type as surface_pressure at end
    type_pressure_level = type(pressure_lev)

    # Preserve attributes for Xarray
    if type_surface_pressure == xr.DataArray:
        da_coords = dict(surface_pressure.coords)
        da_attrs = dict(surface_pressure.attrs)
        da_dims = surface_pressure.dims
    if type_pressure_level == xr.DataArray:
        da_attrs = dict(
            pressure_lev.attrs)  # Overwrite attributes to match pressure_lev

    # Calculate delta pressure
    if np.isscalar(surface_pressure):  # scalar case
        delta_pressure = _delta_pressure1D(pressure_lev, surface_pressure)
    else:  # multi-dimensional cases
        shape = surface_pressure.shape
        delta_pressure_shape = shape + (len(pressure_lev),
                                       )  # preserve shape for reshaping

        surface_pressure_flattened = np.ravel(
            surface_pressure)  # flatten to avoid nested for loops
        delta_pressure = [
            _delta_pressure1D(pressure_lev, e)
            for e in surface_pressure_flattened
        ]

        delta_pressure = np.array(delta_pressure).reshape(delta_pressure_shape)

    # If passed in an Xarray array, return an Xarray array
    # Change this to return a dataset that has both surface pressure and delta pressure?
    if type_surface_pressure == xr.DataArray:
        da_coords['lev'] = pressure_lev.values if (
            type_pressure_level == xr.DataArray) else pressure_lev
        da_dims = da_dims + ("lev",)
        da_attrs.update({"long name": "pressure layer thickness"})
        delta_pressure = xr.DataArray(delta_pressure,
                                      coords=da_coords,
                                      dims=da_dims,
                                      attrs=da_attrs,
                                      name="delta pressure")

    return delta_pressure


# NCL NAME WRAPPER FUNCTIONS BELOW
def dpres_plev(pressure_lev, surface_pressure):
    return delta_pressure(pressure_lev, surface_pressure)


_generate_wrapper_docstring(dpres_plev, delta_pressure)



# move functions into geocat.comp namespace
from .climatologies import anomaly, climatology, month_to_season, calendar_average, climatology_average, climate_anomaly
from .fourier_filters import (fourier_band_block, fourier_band_pass,
                              fourier_filter, fourier_high_pass,
                              fourier_low_pass)
from .gradient import gradient, _arc_lon_wgs84, _arc_lat_wgs84, _rad_lat_wgs84
from .interpolation import interp_hybrid_to_pressure, interp_sigma_to_hybrid, interp_multidim
from .meteorology import (dewtemp, heat_index, relhum, relhum_ice, relhum_water,
                          actual_saturation_vapor_pressure, max_daylight,
                          psychrometric_constant, saturation_vapor_pressure,
                          saturation_vapor_pressure_slope, delta_pressure,
                          dpres_plev)
from .spherical import decomposition, recomposition, scale_voronoi
from .stats import eofunc, eofunc_eofs, eofunc_pcs, eofunc_ts, pearson_r
# bring all functions from geocat.f2py into the geocat.comp namespace
try:
    from geocat.f2py import *
except ImportError:
    pass

# get version from pyproject.toml
from importlib.metadata import version as _version
try:
    __version__ = _version("geocat.comp")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"



from eofs.xarray import Eof
import numpy as np
from typing import Iterable
import xskillscore as xs
import xskillscore.core.np_deterministic as xs_internal
import xarray as xr
import warnings


def pearson_r(a,
              b,
              dim=None,
              weights=None,
              skipna=False,
              keep_attrs=False,
              axis=None):
    """This function wraps the function of the same name from `xskillscore <htt
    ps://xskillscore.readthedocs.io/en/stable/api/xskillscore.pearson_r.html#xs
    killscore.pearson_r>`__. The difference between the xskillscore version and
    this one is that the GeoCAT version allows for array-like inputs rather
    than only supporting ``xarray.DataArrays``. The parameters work the same
    way as in the xskillscore version, with an added parameter ``axis`` (see
    below).

    Parameters
    ----------
    a : :class:`xarray.DataArray`, :class:`xarray.Dataset`, ndarray
        Arrays over which to apply the function.
    b : :class:`xarray.DataArray`, :class:`xarray.Dataset`, ndarray
        Arrays over which to apply the function.
    dim : str, :class:`list`, optional
        The dimension(s) to apply the correlation along. Note that this dimension will be reduced as a result. Defaults to ``None`` reducing all dimensions. Only used when ``a`` and ``b`` are of type ``xarray.DataArray`` or ``xarray.Dataset``.
    weights : :class:`xarray.DataArray`, ndarray, optional
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool, optional
        If True, skip NaNs when computing function.
    keep_attrs : bool, optional
        If True, the attributes (attrs) will be copied from the first input to the new one. If False (default), the new object will be returned without attributes. Only used when ``a`` and ``b`` are of types ``xarray.Dataset`` or ``xarray.DataArray``.
    axis : :class:`int`
        The axis to apply the correlation along. Only used when ``a`` and ``b`` are of type ``np.ndarray`` or are array-like.

    Returns
    -------
    r : :class:`xarray.DataArray`, ndarray
        Pearson's correlation coefficient
    """
    # if a and v are not xr.DataArrays
    if not isinstance(a, xr.DataArray) and not isinstance(b, xr.DataArray):
        if (dim is not None) and (axis is not None):
            warnings.warn(
                "The `dim` keyword is unused with non xarray.DataArray inputs")
        if axis is None:  # squash array to 1D for element wise computation
            axis = 0
            a = np.ravel(a)
            b = np.ravel(b)
            if weights is not None:
                weights = np.ravel(weights)
        try:
            return xs_internal._pearson_r(a, b, weights, axis, skipna)
        except ValueError:
            print('Data along `axis` must have the same dimension as `weights`')
    else:  # if a and b are xr.DataArrays
        if (dim is not None) and (axis is not None):
            warnings.warn(
                "The `axis` keyword is unused with xarray.DataArray inputs")
        try:
            return xs.pearson_r(a, b, dim, weights, skipna, keep_attrs)
        except ValueError:
            print('Data along `dim` must have the same dimension as `weights`')


def _generate_eofs_solver(data, time_dim=0, weights=None, center=True, ddof=1):
    """Convenience function to be used in both `eofunc_eofs` and `eofunc_pcs`
    functions."""

    # ''' Start of boilerplate
    if not isinstance(data, xr.DataArray):

        data = np.asarray(data)

        if (time_dim >= data.ndim) or (time_dim < -data.ndim):
            raise ValueError("ERROR eofunc_efs: `time_dim` out of bound.")

        # Transpose data if time_dim is not 0 (i.e. the first/left-most dimension)
        dims_to_transpose = np.arange(data.ndim).tolist()
        dims_to_transpose.insert(
            0, dims_to_transpose.pop(dims_to_transpose.index(time_dim)))
        data = np.transpose(data, axes=dims_to_transpose)

        dims = [f"dim_{i}" for i in range(data.ndim)]
        dims[0] = 'time'

        data = xr.DataArray(
            data,
            dims=dims,
        )

    solver = Eof(data, weights=weights, center=center, ddof=ddof)

    return data, solver


def eofunc_eofs(data,
                neofs=1,
                time_dim=0,
                eofscaling=0,
                weights=None,
                center=True,
                ddof=1,
                vfscaled=False,
                meta=False):
    """Computes empirical orthogonal functions (EOFs, aka: Principal Component
    Analysis).

    Note: ``eofunc_eofs`` allows to perform the EOF analysis that was previously done
    via the NCL function ``eofunc``.

    However, there are a few changes to the NCL flow such as:

    1. Only ``np.nan`` is supported as missing value,
    2. EOFs are computed only from covariance matrix and there is no support for computation from correlation matrix,
    3. percentage of non-missing points that must exist at any single point is no longer an input.

    This implementation uses the `eofs package <https://anaconda.org/conda-forge/eofs>`__,
    which is built upon the following study: Dawson, Andrew, "eofs: A library
    for EOF analysis of meteorological, oceanographic, and climate data,"
    Journal of Open Research Software, vol. 4, no. 1, 2016. Further information about this
    package can be found `here <https://ajdawson.github.io/eofs/latest/index.html>`__.

    This implementation provides a few conveniences to the user on top of
    ``eofs`` package that are described below.

    Parameters
    ----------
    data : :class:`xarray.DataArray` or ndarray or list
        Should contain numbers or `np.nan` for missing value representation. It must be at least a 2-dimensional array.

        When input data is of type `xarray.DataArray`, `eofs.xarray` interface assumes the left-most dimension
        (i.e. `dim_0`) is the `time` dimension. In this case, that dimension should have the name "time".

        When input data is of type `numpy.ndarray` or `list`, this function still assumes the leftmost dimension
        to be the number of observations or `time` dimension: however, in this case, user is allowed to input otherwise.
        If the input do not have its leftmost dimension as the `time` or number of observations, then the user should
        specify with `time_dim=x` to define which dimension must be treated as time or number of observations

    neofs : int, optional
        A scalar integer that specifies the number of empirical orthogonal functions (i.e. eigenvalues and
        eigenvectors) to be returned. This is usually less than or equal to the minimum number of observations or
        number of variables. Defaults to 1.

    time_dim : int, optional
        An integer defining the time dimension if it is not the leftmost dimension. When input data is of type
        `xarray.DataArray`, this is ignored (assuming `xarray.DataArray` has its leftmost dimension with the exact
        name 'time'). It must be between ``0`` and ``data.ndim - 1`` or it could be ``-1`` indicating the last
        dimension. Defaults to 0.

        Note: The `time_dim` argument allows to perform the EOF analysis that was previously done via the NCL
        function `eofunc_n`.

    eofscaling : int, optional
        (From `eofs` package): Sets the scaling of the EOFs. The following values are accepted:

            - 0 : Un-scaled EOFs (default).
            - 1 : EOFs are divided by the square-root of their eigenvalues.
            - 2 : EOFs are multiplied by the square-root of their eigenvalues.

    weights : array_like, optional
        (From `eofs` package): An array of weights whose shape is compatible with those of the input array dataset.
        The weights can have the same shape as dataset or a shape compatible with an array broadcast (i.e., the shape
        of the weights can can match the rightmost parts of the shape of the input array dataset). If the input array
        dataset does not require weighting then the value None may be used. Defaults to ``None`` (no weighting).

    center : bool, optional
        (From `eofs` package): If True, the mean along the first axis of dataset (the time-mean) will be removed prior
        to analysis. If False, the mean along the first axis will not be removed. Defaults to True (mean is removed).

        The covariance interpretation relies on the input data being anomaly data with a time-mean of 0. Therefore this
        option should usually be set to True. Setting this option to True has the useful side effect of propagating
        missing values along the time dimension, ensuring that a solution can be found even if missing values occur
        in different locations at different times.

    ddof : int, optional
        (From `eofs` package): ‘Delta degrees of freedom’. The divisor used to normalize the covariance matrix is
        N - ddof where N is the number of samples. Defaults to 1.

    vfscaled : bool, optional
        (From `eofs` package): If True, scale the errors by the sum of the eigenvalues. This yields typical errors
        with the same scale as the values returned by Eof.varianceFraction. If False then no scaling is done.
        Defaults to False.

    meta : bool, optional
        If set to True and the input array is an Xarray, the metadata from the input array will be copied to the
        output array. Defaults to False.

    Returns
    -------
        A multi-dimensional array containing EOFs. The returned array will be of the same size as data with the
        leftmost dimension removed and an additional dimension of the size `neofs` added.

        The return variable will have associated with it the following attributes:

        eigenvalues:
            A one-dimensional array of size `neofs` that contains the eigenvalues associated with each EOF.

        northTest:
            (From `eofs` package): Typical errors for eigenvalues.

            The method of North et al. (1982) is used to compute the typical error for each eigenvalue. It is
            assumed that the number of times in the input data set is the same as the number of independent
            realizations. If this assumption is not valid then the result may be inappropriate.

            Note: The `northTest` attribute allows to perform the error analysis that was previously done via the NCL
            function `eofunc_north`.

        totalAnomalyVariance:
            (From `eofs` package): Total variance associated with the field of anomalies (the sum of the eigenvalues).

        varianceFraction:
            (From `eofs` package): Fractional EOF mode variances.

            The fraction of the total variance explained by each EOF mode, values between 0 and 1 inclusive..

    See Also
    --------
    Related NCL Functions:
    `eofunc <https://www.ncl.ucar.edu/Document/Functions/Built-in/eofunc.shtml>`__,
    `eofunc_Wrap <https://www.ncl.ucar.edu/Document/Functions/Contributed/eofunc_Wrap.shtml>`__,
    `eofunc_north <https://www.ncl.ucar.edu/Document/Functions/Contributed/eofunc_north.shtml>`__,
    `eofunc_n <https://www.ncl.ucar.edu/Document/Functions/Built-in/eofunc_n.shtml>`__,
    `eofunc_n_Wrap <https://www.ncl.ucar.edu/Document/Functions/Contributed/eofunc_n_Wrap.shtml>`__
    """
    data, solver = _generate_eofs_solver(data,
                                         time_dim=time_dim,
                                         weights=weights,
                                         center=center,
                                         ddof=ddof)

    # Checking number of EOFs
    if neofs <= 0:
        raise ValueError(
            "ERROR eofunc_eofs: num_eofs must be a positive non-zero integer value."
        )

    eofs = solver.eofs(neofs=neofs, eofscaling=eofscaling)

    # Populate attributes for output
    attrs = {}

    if meta:
        attrs = data.attrs

    attrs['eigenvalues'] = solver.eigenvalues(neigs=neofs)
    attrs['northTest'] = solver.northTest(neigs=neofs, vfscaled=vfscaled)
    attrs['totalAnomalyVariance'] = solver.totalAnomalyVariance()
    attrs['varianceFraction'] = solver.varianceFraction(neigs=neofs)

    if meta:
        dims = ["eof"
               ] + [data.dims[i] for i in range(data.ndim) if i != time_dim]
        coords = {
            k: v for (k, v) in data.coords.items() if k != data.dims[time_dim]
        }
    else:
        dims = ["eof"] + [f"dim_{i}" for i in range(data.ndim) if i != time_dim]
        coords = {}

    return xr.DataArray(eofs, attrs=attrs, dims=dims, coords=coords)


def eofunc_pcs(data,
               npcs=1,
               time_dim=0,
               pcscaling=0,
               weights=None,
               center=True,
               ddof=1,
               meta=False):
    """Computes the principal components (time projection) in the empirical
    orthogonal function analysis.

    Note: ``eofunc_pcs`` allows to perform the analysis that was previously done via the NCL function ``eofunc_ts``.

    However, there are a few changes to the NCL flow such as:

    1. Only `np.nan` is supported as missing value,
    2. EOFs are computed only from covariance matrix and there is no support for computation from correlation matrix,
    3. percentage of non-missing points that must exist at any single point is no longer an input.

    This implementation uses `eofs package <https://anaconda.org/conda-forge/eofs>`__, which is built upon the
    following study: Dawson, Andrew, "eofs: A library for EOF analysis of meteorological, oceanographic, and
    climate data," Journal of Open Research Software, vol. 4, no. 1, 2016. Further information about this
    package can be found `here <https://ajdawson.github.io/eofs/latest/index.html>`__.

    This implementation provides a few conveniences to the user on top of ``eofs`` package that are described below
    in the Parameters section.

    Parameters
    ----------
    data : :class:`xarray.DataArray` or ndarray or list
        Should contain numbers or ``np.nan`` for missing value representation. It must be at least a 2-dimensional array.

        When input data is of type ``xarray.DataArray``, ``eofs.xarray`` interface assumes the left-most dimension
        (i.e. ``dim_0``) is the time dimension. In this case, that dimension should have the name "time".

        When input data is of type ``numpy.ndarray`` or ``list``, this function still assumes the leftmost dimension
        to be the number of observations or time dimension: however, in this case, user is allowed to input otherwise.
        If the input do not have its leftmost dimension as the time or number of observations, then the user should
        specify with ``time_dim=x`` to define which dimension must be treated as time or number of observations

    npcs : int, optional
        A scalar integer that specifies the number of principal components (i.e. eigenvalues and eigenvectors) to be
        returned. This is usually less than or equal to the minimum number of observations or number of variables.
        Defaults to 1.

    time_dim : int, optional
        An integer defining the time dimension if it is not the leftmost dimension. When input data is of type
        `xarray.DataArray`, this is ignored (assuming `xarray.DataArray` has its leftmost dimension with the exact
        name 'time'). It must be between ``0`` and ``data.ndim - 1`` or it could be ``-1`` indicating the last
        dimension. Defaults to 0.

        Note: The `time_dim` argument allows to perform the EOF analysis that was previously done via the NCL
        function `eofunc_ts_n`.

    pcscaling : int, optional
        (From ``eofs`` package): Sets the scaling of the retrieved PCs. The following values are accepted:
            - 0 : Un-scaled PCs (default).
            - 1 : PCs are divided by the square-root of their eigenvalues.
            - 2 : PCs are multiplied by the square-root of their eigenvalues.

    weights : array_like, optional
        (From ``eofs`` package): An array of weights whose shape is compatible with those of the input array dataset.
        The weights can have the same shape as dataset or a shape compatible with an array broadcast (i.e., the shape
        of the weights can can match the rightmost parts of the shape of the input array dataset). If the input array
        dataset does not require weighting then the value None may be used. Defaults to ``None`` (no weighting).

    center : bool, optional
        (From ``eofs`` package): If True, the mean along the first axis of dataset (the time-mean) will be removed prior
        to analysis. If False, the mean along the first axis will not be removed. Defaults to True (mean is removed).

        The covariance interpretation relies on the input data being anomaly data with a time-mean of 0. Therefore this
        option should usually be set to True. Setting this option to True has the useful side effect of propagating
        missing values along the time dimension, ensuring that a solution can be found even if missing values occur
        in different locations at different times.

    ddof : int, optional
        (From ``eofs`` package): ‘Delta degrees of freedom’. The divisor used to normalize the covariance matrix is
        N - ddof where N is the number of samples. Defaults to 1.

    meta : bool, optional
        If set to ``True`` and the input array is an Xarray, the metadata from the input array will be copied to the
        output array. Defaults to False.

    Returns
    -------

    See Also
    --------
    Related NCL Functions:
    `eofunc_ts <https://www.ncl.ucar.edu/Document/Functions/Built-in/eofunc_ts.shtml>`__,
    `eofunc_ts_Wrap <https://www.ncl.ucar.edu/Document/Functions/Contributed/eofunc_ts_Wrap.shtml>`__,
    `eofunc_ts_n <https://www.ncl.ucar.edu/Document/Functions/Built-in/eofunc_ts_n.shtml>`__,
    `eofunc_ts_n_Wrap <https://www.ncl.ucar.edu/Document/Functions/Contributed/eofunc_ts_n_Wrap.shtml>`__
    """

    data, solver = _generate_eofs_solver(data,
                                         time_dim=time_dim,
                                         weights=weights,
                                         center=center,
                                         ddof=ddof)

    # Checking number of EOFs
    if npcs <= 0:
        raise ValueError(
            "ERROR eofunc_pcs: num_pcs must be a positive non-zero integer value."
        )

    solver = Eof(data, weights=weights, center=center, ddof=ddof)

    pcs = solver.pcs(npcs=npcs, pcscaling=pcscaling)
    pcs = pcs.transpose()

    # Populate attributes for output
    attrs = {}

    if meta:
        attrs = data.attrs

    dims = ["pc", "time"]
    if meta:
        coords = {"time": data.coords[data.dims[time_dim]]}
    else:
        coords = {}

    return xr.DataArray(pcs, attrs=attrs, dims=dims, coords=coords)


# Transparent wrappers for geocat.comp backwards compatibility


def eofunc(data: Iterable, neval, **kwargs) -> xr.DataArray:
    warnings.warn(
        "eofunc will be deprecated soon in a future version and may not currently generate proper results for some of "
        "its arguments including `pcrit`, `jopt="
        "correlation"
        "`, and 'missing_value' other than np.nan. The output "
        " and its attributes may thus not be as expected, too. Use `eofunc_eofs` instead.",
        PendingDeprecationWarning)

    if not isinstance(data, xr.DataArray) or not isinstance(data, np.ndarray):
        data = np.asarray(data)

    time_dim = int(kwargs.get("time_dim", data.ndim - 1))
    meta = bool(kwargs.get("meta"))

    return eofunc_eofs(data, neofs=neval, time_dim=time_dim, meta=meta)


def eofunc_ts(data: Iterable, evec, **kwargs) -> xr.DataArray:
    warnings.warn(
        "eofunc_ts will be deprecated soon in a future version and may not currently generate proper results for "
        "some of its arguments including `evec`, `jopt="
        "correlation"
        "`, and 'missing_value' other than np.nan. The output "
        " and its attributes may thus not be as expected, too. Use `eofunc_pcs` instead.",
        PendingDeprecationWarning)

    if not isinstance(data, xr.DataArray) or not isinstance(data, np.ndarray):
        data = np.asarray(data)

    time_dim = int(kwargs.get("time_dim", data.ndim - 1))
    meta = bool(kwargs.get("meta"))

    return eofunc_pcs(data, npcs=evec.shape[0], time_dim=time_dim, meta=meta)



from typing import Union

import numpy as np
import scipy.special as sspecial
import scipy.spatial as sspatial
import xarray as xr

SupportedTypes = Union[np.ndarray, xr.DataArray]
default_max_harm = 23  # 300 harmonics from 0,0 to 23,23


def decomposition(
    data: SupportedTypes,
    scale: SupportedTypes,
    theta: SupportedTypes,
    phi: SupportedTypes,
    max_harm: int = default_max_harm,
    chunk_size: int = 'auto',
) -> SupportedTypes:
    """Calculate the spherical harmonics of a dataset.

    Parameters
    ----------
    data : ndarray, :class:`xarray.DataArray`
        2-dimensional dataset

    scale : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the weighting of each point in the data.
        This is usually the area of the voronoi cell centered on the corresponding datapoint.
        the ``geocat.comp.spherical.scale_voronoi(theta,phi)`` function can provide a scale array.

    theta : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the theta (longitude in radians) values for each datapoint in data.

    phi : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the ``theta`` (latitude in radians) values for each datapoint in data.
        ``Phi`` is zero at the top of the sphere and ``pi`` at the bottom, ``phi = (lat_degrees-90)*(-1)*pi/180``

    max_harm: int, optional
        The maximum harmonic value for both m and n.
        The total of harmonics calculated is ``(max_harm+1)*(max_harm+2)/2``
        Defaults to 23, for 300 total harmonics.

    chunk_size: int, optional
        The size of each edge of the dask chunks if using xarray.DataArray inputs.
        Some arrays will be 2d, and others 1d, and the final calculation operates on a 3d array.
        thus the chunks used in the largest calculation scale at ``chunk_size^3``
        A chunk size of 256 is recommended. Defaults to 'auto'

    Returns
    -------
    decomposition : ndarray, :class:`xarray.DataArray`
        The spherical harmonic decomposition of the input data
    """

    # scale_val is the inverse of the total sphere area times the magnitude of
    # the first harmonic. This is used to scale the output so that the output
    # is unaffected by the surface area of the original sphere.
    scale_val = 1 / (np.sum(scale, axis=(0, 1)) *
                     sspecial.sph_harm(0, 0, 0, 0)**2)

    mlist = []  # ordered list of the m harmonics sspecial.sphere(m,n,theta,phi)
    nlist = []  # ordered list of the n harmonics sspecial.sphere(m,n,theta,phi)
    # the real value of the output varies on a factor of two when m is 0,
    # due to m=0 being a symmetric case with no imaginary component,
    # this accounts for that difference in output sum.
    scale_mul = []
    for nvalue in range(max_harm + 1):
        for mvalue in range(nvalue + 1):
            mlist.append(mvalue)
            nlist.append(nvalue)
            if mvalue == 0:
                scale_mul.append(1)
            else:
                scale_mul.append(2)
    m = np.array(mlist)
    n = np.array(nlist)
    scale_mul = np.array(scale_mul)

    # if numpy, change dimensions to allow for broadcast in ss.sph_harm
    if type(data) is np.ndarray:
        m = np.expand_dims(m, axis=(0, 1))
        n = np.expand_dims(n, axis=(0, 1))
        theta = np.expand_dims(theta, axis=(2))
        phi = np.expand_dims(phi, axis=(2))
        scale_res = scale_mul * scale_val
        scale_dat = np.expand_dims(np.multiply(data, scale), axis=(2))

    # if xarray, set dims and chunks for broadcast in ss.sphere_harm
    if type(data) is xr.DataArray:
        m = xr.DataArray(m, dims=['har']).chunk((chunk_size))
        n = xr.DataArray(n, dims=['har']).chunk((chunk_size))
        scale_res = xr.DataArray(
            scale_mul,
            dims=['har'],
        ).chunk((chunk_size)) * scale_val
        scale_dat = xr.DataArray(
            np.multiply(data, scale),
            dims=data.dims,
        ).chunk((chunk_size))
        theta = xr.DataArray(theta, dims=data.dims).chunk((chunk_size))
        phi = xr.DataArray(phi, dims=data.dims).chunk((chunk_size))

    results = np.sum(
        np.multiply(scale_dat, sspecial.sph_harm(m, n, theta, phi)),
        axis=(0, 1),
    ) * scale_res
    return results


def recomposition(
    data: SupportedTypes,
    theta: SupportedTypes,
    phi: SupportedTypes,
    max_harm: int = default_max_harm,
    chunk_size: int = 'auto',
) -> SupportedTypes:
    """Calculate a dataset from spherical harmonics.

    Parameters
    ----------
    data : ndarray, :class:`xarray.DataArray`
        1-dimensional array of spherical harmonics.
        These must be in the same order output by ``geocat.comp.spherical.decomposition``.

    theta : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the theta (longitude in radians) values for each datapoint in data.

    phi : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the theta (latitude in radians) values for each datapoint in data.
        ``Phi`` is zero at the top of the sphere and ``pi`` at the bottom, ``phi = (lat_degrees-90)*(-1)*pi/180``

    max_harm: int, optional
        The maximum harmonic value for both m and n.
        The total of harmonics calculated is ``(max_harm+1)*(max_harm+2)/2``
        The number of total harmonics must equal the number of harmoncs in the input data.
        Defaults to 23, for 300 total harmonics.

    chunk_size: int, optional
        The size of each edge of the dask chunks if using xarray.DataArray inputs.
        Some arrays will be 2d, and others 1d, and the final calculation operates on a 3d array.
        thus the chunks used in the largest calculation scale at ``chunk_size^3``
        A chunk size of 256 is recommended. Defaults to 'auto'

    Returns
    -------
    recomposition : ndarray, :class:`xarray.DataArray`
        The spherical harmonic recomposition of the input data
    """

    mlist = []  # ordered list of the m harmonics sspecial.sphere(m,n,theta,phi)
    nlist = []  # ordered list of the n harmonics sspecial.sphere(m,n,theta,phi)
    for nvalue in range(max_harm + 1):
        for mvalue in range(nvalue + 1):
            mlist.append(mvalue)
            nlist.append(nvalue)
    m = np.array(mlist)
    n = np.array(nlist)

    # if numpy, change dimensions to allow for broadcast in ss.sph_harm
    if type(data) is np.ndarray:
        m = np.expand_dims(m, axis=(1, 2))
        n = np.expand_dims(n, axis=(1, 2))
        data = np.expand_dims(data, axis=(1, 2))
        theta = np.expand_dims(theta, axis=(0))
        phi = np.expand_dims(phi, axis=(0))

    # if xarray, set dims and chunks for broadcast in ss.sphere_harm
    if type(data) is xr.DataArray:
        m = xr.DataArray(m, dims=['har']).chunk((chunk_size))
        n = xr.DataArray(n, dims=['har']).chunk((chunk_size))
        data = xr.DataArray(data, dims=['har']).chunk((chunk_size))
        theta = xr.DataArray(theta, dims=theta.dims).chunk((chunk_size))
        phi = xr.DataArray(phi, dims=phi.dims).chunk((chunk_size))

    results = np.sum(
        np.multiply(sspecial.sph_harm(m, n, theta, phi).real, data.real),
        axis=(0),
    ) + np.sum(
        np.multiply(sspecial.sph_harm(m, n, theta, phi).imag, data.imag),
        axis=(0),
    )

    return results.real


def scale_voronoi(
    theta: SupportedTypes,
    phi: SupportedTypes,
    chunk_size: int = 'auto',
) -> SupportedTypes:
    """Calculate the area weighting for dataset.

    Parameters
    ----------
    theta : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the theta (longitude in radians) values for each datapoint in data.

    phi : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the theta (latitude in radians) values for each datapoint in data.
        ``Phi`` is zero at the top of the sphere and ``pi`` at the bottom, ``phi = (lat_degrees-90)*(-1)*pi/180``

    chunk_size: int, optional
        The size of each edge of the dask chunks if using xarray.DataArray inputs.
        Some arrays will be 2d, and others 1d, and the final calculation operates on a 3d array.
        thus the chunks used in the largest calculation scale at ``chunk_size^3``
        A chunk size of 256 is recommended. Defaults to 'auto'

    Returns
    -------
    scale : ndarray, :class:`xarray.DataArray`
        2-dimensional array containing the area of the spherical voronoi cell for each ``theta`` and ``phi`` pair.
    """

    if type(theta) is xr.DataArray:
        theta = theta.to_numpy()
        phi = phi.to_numpy()

    theta_1d = theta.reshape((theta.shape[0] * theta.shape[1],))
    phi_1d = phi.reshape((phi.shape[0] * phi.shape[1],))

    data_locs_3d = np.zeros((len(phi_1d), 3))
    data_locs_3d[:, 0] = np.sin(phi_1d) * np.sin(theta_1d)
    data_locs_3d[:, 1] = np.sin(phi_1d) * np.cos(theta_1d)
    data_locs_3d[:, 2] = np.cos(phi_1d)

    scale = np.array(
        sspatial.SphericalVoronoi(
            data_locs_3d,
            radius=1.0,
            center=np.array([0, 0, 0]),
        ).calculate_areas()).reshape(theta.shape)

    if type(theta) is xr.DataArray:
        scale = xr.DataArray(scale, dims=theta.dims).chunk(chunk_size)

    return scale



import math as m
import numpy as np
import typing
import xarray as xr


def fourier_filter(
        signal: typing.Union[np.ndarray, xr.DataArray],
        frequency: float,
        cutoff_frequency_low: float = 0,
        cutoff_frequency_high: float = 0,
        time_axis: int = 0,
        low_pass: bool = False,
        high_pass: bool = False,
        band_pass: bool = False,
        band_block: bool = False) -> typing.Union[np.ndarray, xr.DataArray]:
    """Filter a dataset by frequency. This function allows for low_pass,
    high_pass, band_pass, or band_block filtering of the data's freqency
    representation.

    Parameters
    ----------
    signal : ndarray, :class:`xarray.DataArray`
        n-dimensional dataset

    frequency : :class:`float`
        sample frequency of dataset

    cutoff_frequency_low : float, optional
        low frequency for cutting fourier transform, used by low_pass, band_pass, band_block. Defaults to 0.

    cutoff_frequency_high : float, optional
        high frequency for cutting fourier transform, used by low_pass, band_pass, band_block. Defaults to 0.

    time_axis : int, optional
        the time axis of the data set. Defaults to 0.

    low_pass : bool, optional
        runs a low_pass filter on the data if set to True. Defaults to False.

    high_pass : bool, optional
        runs a high_pass filter on the data if set to True. Defaults to False.

    band_pass : bool, optional
        runs a band_pass filter on the data if set to True. Defaults to False.

    band_block : bool, optional
        runs a band_block filter on the data if set to True. Defaults to False.

    Returns
    -------
    return_signal : ndarray, :class:`xarray.DataArray`
        signal with specified filters applied

    Examples
    --------
    Example 1: The tidal cycle needs to be removed from a 10/hr `oceanic dataset <https://tidesandcurrents.noaa.gov/waterlevels.html?id=9415020&units=standard&bdate=20210101&edate=20210131&timezone=GMT&datum=MLLW&interval=6&action=data>`__


    >>> from geocat.comp import fourier_filter
    >>> import matplotlib.pyplot as plt
    >>> from mpl_toolkits.mplot3d import Axes3D
    >>> import numpy as np
    >>> import pandas as pd
    >>> import xarray as xr

    >>> dataset = xr.DataArray(pd.read_csv('CO-OPS_9415020_wl.csv'))
    >>> xr_data = dataset.loc[:,'Verified (ft)']

    >>> data_freq = 10 #points per hour
    >>> tide_freq1 = 1/(1*12.4206) #tides per hour
    >>> tide_freq2 = 1/(2*12.4206) #tides per hour
    >>> res = data_freq/(len(xr_data))
    >>> cflow1 = tide_freq1 - res*5
    >>> cfhigh1 = tide_freq1 + res*5
    >>> cflow2 = tide_freq2 - res*5
    >>> cfhigh2 = tide_freq2 + res*5

    >>> fig, ax = plt.subplots(1,
    ...                        1,
    ...                        dpi=100,
    ...                        figsize=(8,4),
    ...                        constrained_layout=True)
    >>> no_tide = xr_data
    >>> ax.plot(no_tide[2000:3000])
    >>> no_tide = fourier_filter(no_tide,
    ...                          data_freq,
    ...                          cutoff_frequency_low=cflow1,
    ...                          cutoff_frequency_high=cfhigh1,
    ...                          band_block=True)
    >>> ax.plot(no_tide[2000:3000])
    >>> no_tide = fourier_filter(no_tide,
    ...                          data_freq,
    ...                          cutoff_frequency_low=cflow2,
    ...                          cutoff_frequency_high=cfhigh2,
    ...                          band_block=True)
    >>> ax.plot(no_tide[2000:3000])
    >>> fig.show()

    >>> fig, axs = plt.subplots(2,
    ...                         1,
    ...                         dpi=100,
    ...                         figsize=(8,4),
    ...                         constrained_layout=True)
    >>> axs[0].set_title('real')
    >>> axs[0].plot(np.real(np.fft.fft(xr_data)[1:100]))
    >>> axs[0].plot(np.real(np.fft.fft(no_tide)[1:100]))
    >>> axs[1].set_title('imag')
    >>> axs[1].plot(np.imag(np.fft.fft(xr_data)[1:100]))
    >>> axs[1].plot(np.imag(np.fft.fft(no_tide)[1:100]))
    >>> fig.show()

    >>> fig, axs = plt.subplots(2,
    ...                         1,
    ...                         dpi=100,
    ...                         figsize=(8,4),
    ...                         constrained_layout=True)
    >>> start = 0
    >>> end = -1
    >>> axs[0].set_title('real')
    >>> axs[0].plot(np.real(xr_data)[start:end])
    >>> axs[0].plot(np.real(no_tide)[start:end])
    >>> axs[1].set_title('imag')
    >>> axs[1].plot(np.imag(xr_data)[start:end])
    >>> axs[1].plot(np.imag(no_tide)[start:end])
    >>> fig.show()
    """
    resolution = frequency / len(signal)
    signal = np.swapaxes(signal, time_axis, 0)
    res_fft = np.fft.fft(signal, axis=0)
    cfl_index = m.floor(cutoff_frequency_low / resolution)
    cfln_index = 1 - cfl_index
    cfh_index = m.ceil(cutoff_frequency_high / resolution)
    cfhn_index = 1 - cfh_index
    if low_pass:
        if cfl_index > 1:
            res_fft[cfl_index:cfln_index] = np.zeros(
                res_fft[cfl_index:cfln_index].shape)
        else:
            res_fft[cfl_index:] = np.zeros(res_fft[cfl_index:].shape)
    if high_pass:
        res_fft[:cfh_index] = np.zeros(res_fft[:cfh_index].shape)
        if cfh_index > 1:
            res_fft[cfhn_index:] = np.zeros(res_fft[cfhn_index:].shape)
    if band_pass:
        res_fft[:cfl_index] = np.zeros(res_fft[:cfl_index].shape)
        if cfh_index > 1:
            res_fft[cfh_index:cfhn_index] = np.zeros(
                res_fft[cfh_index:cfhn_index].shape)
        else:
            res_fft[cfh_index:] = np.zeros(res_fft[cfh_index:].shape)
        if cfl_index > 1:
            res_fft[cfln_index:] = np.zeros(res_fft[cfln_index:].shape)
    if band_block:
        res_fft[cfl_index:cfh_index] = np.zeros(
            res_fft[cfl_index:cfh_index].shape)
        if cfl_index > 1 and cfh_index > 1:
            res_fft[cfhn_index:cfln_index] = np.zeros(
                res_fft[cfhn_index:cfln_index].shape)
        elif cfh_index > 1:
            res_fft[cfhn_index:] = np.zeros(res_fft[cfhn_index:].shape)
    result = np.fft.ifft(res_fft, axis=0)
    result = np.real(result)
    result = np.swapaxes(result, time_axis, 0)
    if type(signal) == xr.DataArray:
        xr_result = signal.copy()
        xr_result.data = result
        result = xr_result
    return result


def fourier_low_pass(
        signal: typing.Union[np.ndarray, xr.DataArray],
        frequency: float,
        cutoff_frequency_low: float,
        time_axis: int = 0) -> typing.Union[np.ndarray, xr.DataArray]:
    """Filter a dataset by frequency. This function allowes for low_pass
    filtering of the data's freqency representation.

    Parameters
    ----------
    signal : ndarray, :class:`xarray.DataArray`
        n-dimensional dataset

    frequency : :class:`float`
        sample frequency of dataset

    cutoff_frequency_low : :class:`float`
        low frequency for cutting fourier transform

    time_axis : int, optional
        the time axis of the data set. Defaults to 0.

    Returns
    -------
    return_signal : ndarray, :class:`xarray.DataArray`
        signal with specified filters applied
    """
    return fourier_filter(signal,
                          frequency,
                          cutoff_frequency_low=cutoff_frequency_low,
                          time_axis=time_axis,
                          low_pass=True)


def fourier_high_pass(
        signal: typing.Union[np.ndarray, xr.DataArray],
        frequency: float,
        cutoff_frequency_high: float,
        time_axis: int = 0) -> typing.Union[np.ndarray, xr.DataArray]:
    """Filter a dataset by frequency. This function allowes for high_pass
    filtering of the data's freqency representation.

    Parameters
    ----------
    signal : ndarray, :class:`xarray.DataArray`
        n-dimensional dataset

    frequency : :class:`float`
        sample frequency of dataset

    cutoff_frequency_high : :class:`float`
        high frequency for cutting fourier transform

    time_axis : int, optional
        the time axis of the data set. Defaults to 0.

    Returns
    -------
    return_signal : ndarray, :class:`xarray.DataArray`
        signal with specified filters applied
    """
    return fourier_filter(signal,
                          frequency,
                          cutoff_frequency_high=cutoff_frequency_high,
                          time_axis=time_axis,
                          high_pass=True)


def fourier_band_pass(
        signal: typing.Union[np.ndarray, xr.DataArray],
        frequency: float,
        cutoff_frequency_low: float,
        cutoff_frequency_high: float,
        time_axis: int = 0) -> typing.Union[np.ndarray, xr.DataArray]:
    """Filter a dataset by frequency. This function allowes for band_pass
    filtering of the data's freqency representation.

    Parameters
    ----------
    signal : ndarray, :class:`xarray.DataArray`
        n-dimensional dataset

    frequency : :class:`float`
        sample frequency of dataset

    cutoff_frequency_low : :class:`float`
        low frequency for cutting fourier transform

    cutoff_frequency_high : :class:`float`
        high frequency for cutting fourier transform

    time_axis : int, optional
        the time axis of the data set. Defaults to 0.

    Returns
    -------
    return_signal : ndarray, :class:`xarray.DataArray`
        signal with specified filters applied
    """
    return fourier_filter(signal,
                          frequency,
                          cutoff_frequency_low=cutoff_frequency_low,
                          cutoff_frequency_high=cutoff_frequency_high,
                          time_axis=time_axis,
                          band_pass=True)


def fourier_band_block(
        signal: typing.Union[np.ndarray, xr.DataArray],
        frequency: float,
        cutoff_frequency_low: float,
        cutoff_frequency_high: float,
        time_axis: int = 0) -> typing.Union[np.ndarray, xr.DataArray]:
    """Filter a dataset by frequency. This function allowes for band_block
    filtering of the data's freqency representation.

    Parameters
    ----------
    signal : ndarray, :class:`xarray.DataArray`
        n-dimensional dataset

    frequency : :class:`float`
        sample frequency of dataset

    cutoff_frequency_low : :class:`float`
        low frequency for cutting fourier transform

    cutoff_frequency_high : :class:`float`
        high frequency for cutting fourier transform

    time_axis : int, optional
        the time axis of the data set. Defaults to 0.

    Returns
    -------
    return_signal : ndarray, :class:`xarray.DataArray`
        signal with specified filters applied
    """
    return fourier_filter(signal,
                          frequency,
                          cutoff_frequency_low=cutoff_frequency_low,
                          cutoff_frequency_high=cutoff_frequency_high,
                          time_axis=time_axis,
                          band_block=True)
